#include <stdio.h>
#include <stdlib.h>

#include <velib/velib_inc_mk2.h>
#include <velib/base/base.h>
#include <velib/base/endian.h>
#include <velib/mk2/frame_handler.h>
#include <velib/mk2/mk2.h>
#include <velib/mk2/mk2_update.h>
#include <velib/mk2/vebus_device.h>
#include <velib/utils/ve_assert.h>
#include <velib/utils/ve_logger.h>
#include <velib/mk2/vebus_system.h>

#include "vebusfr_address_discovery.h"
#include "mk2_init.h"

#define DEFAULT_MASTER_ADDR 0

static VebusVersion masterVersion;

static void mk2InitLongWinmonFrSuppDone(void)
{
	if (vebusSystem.init.cb) {
		Mk2Callback* cb = vebusSystem.init.cb;
		vebusSystem.init.cb = NULL;
		cb();
	}
}

static void mk2CheckWinmonFrameVersionDone(void)
{
	if ((masterVersion.features & VEBUS_VERION_F_COMBINED_FRAME) == 0) {
		logI("mk2_init", "Using old type Winmon frames");
		mk2SetState(&mk2.state, mk2InitLongWinmonFrSuppDone); /* Change back to old Winmon frames */
		return;
	}

	logI("mk2_init", "Using new type long Winmon frames");
	mk2InitLongWinmonFrSuppDone();
}

static void mk2SendSetWinmonFrameStateDone(void)
{
	struct VebusDeviceS* dev;
	struct VebusDeviceS mDev;

	mDev.addr = DEFAULT_MASTER_ADDR;

	dev = vebusSystem.devices;
	if (dev == NULL)
		dev = &mDev;			/* No known devices, try master address */

	/* Get the master version to check which version Winmon frames are supported. */
	vebusGetTargetVersion(dev, mk2CheckWinmonFrameVersionDone, & masterVersion);
}

static void mk2VebusDiscoverDoneCb (void)
{
	VebusDevice* dev;

	logI("mk2_init", "mk2VebusDiscoverDoneCb");

	/* Check if the system is properly configured
	 * Unique short addresses */

	vebusSystem.state.shortIds = 0;

	dev = vebusSystem.devices;

	/* Report the connected devices to the main application */
	while (dev != NULL) {
		vebusSystem.state.shortIds |= (1 << dev->addr);
		mk2NewDeviceFound(dev);
		dev = dev->next;
	}

	vebusSystem.state.knownIds = vebusSystem.state.shortIds;

	if (vebusSystem.init.cb) {
		Mk2Callback* cb = vebusSystem.init.cb;
		vebusSystem.init.cb = NULL;
		cb();
	}
}

#if CFG_WITH_MK2_UPDATE
static void stateForUpdateSet(void)
{
	mk2FirmwareUpgrade(vebusSystem.mk2Interface.firmwUpgCb);
}
#endif

static void mk2GetVersionDone(void)
{
	if ((mk2.version % 1000) < 134) {
		logE("mk2_init", "mk2 firmware is too old, upgrade to at least 134");

#if CFG_WITH_MK2_UPDATE
		vebusSystem.init.cb = NULL; /* We will not callback. Call mk2UpdateDoneCb instead */
		mk2.state.sFlags = MK2_S_FLAG_RESET_ON_BREAK | MK2_S_FLAG_DO_NOT_SEND_STATE;
		mk2SetState(&mk2.state, stateForUpdateSet);
		return;
#else
		if (vebusSystem.mk2Interface.firmwUpgCb != NULL)
			vebusSystem.mk2Interface.firmwUpgCb(MK2_UPDATE_FAILED);

		return;
#endif
	}

	vebusDiscoverSystem(&vebusSystem, mk2VebusDiscoverDoneCb);
}


static void mk2CheckVersion(void)
{
	/* Check if the mk2 needs to be updated */
	mk2GetVersion(&mk2.version, mk2GetVersionDone);
}

/**
 * @brief First step: lets see if the mk2 found a vebus device,
 *        if not it needs to be restarted
 *
 * @param cb				Callback when done.
 */
void mk2Start(Mk2Callback* cb)
{
	veAssert(vebusSystem.init.cb == NULL);

	vebusSystem.init.cb = cb;
	vebusDiscoverSystem(&vebusSystem, mk2VebusDiscoverDoneCb);
}

/**
 * @brief First step: Check the mk2frimware version and
 *                    lets see if the mk2 found a vebus device,
 *                    if not it needs to be restarted
 *
 * @param cb				Callback when done.
 * @param cbUpg				Callback after mk2 firmware update
 */
void mk2StartWithFwCheck(Mk2Callback* cb, Mk2ResultCallback* cbUpg)
{
	veAssert(vebusSystem.init.cb == NULL);

	vebusSystem.init.cb = cb;
	vebusSystem.mk2Interface.firmwUpgCb = cbUpg;

	mk2.state.sFlags = MK2_S_FLAG_AUTO_FORWARD_PANEL_FRAMES | MK2_S_FLAG_DO_NOT_SEND_STATE;
	mk2.state.sFlagsExt0 = 1; /* Start with old winmon frames */
	mk2SetState(&mk2.state, mk2CheckVersion);
}

/**
 * @brief Init long Winmon frame support
 *
 * Long Winmon frames are required to retrieve
 * assistant info from the version frame.
 * If not supported this will revert back to the
 * older Winmon frames
 *
 * @param cb				Callback when done.
 */
void mk2InitLongWinmonFrSupp(Mk2Callback* cb)
{
	veAssert(vebusSystem.init.cb == NULL);
	logI("mk2_init", "Init long Winmon frame support");

	vebusSystem.init.cb = cb;

	mk2.state.sFlags = MK2_S_FLAG_AUTO_FORWARD_PANEL_FRAMES | MK2_S_FLAG_DO_NOT_SEND_STATE;
	mk2.state.sFlagsExt0 = 0; /* Use new long winmon frames */
	mk2SetState(&mk2.state, mk2SendSetWinmonFrameStateDone);
}
