/**
 * Successive approximation search functions
 * Search the index of the highest valid object.
 * Valid objects should form a list without gaps starting at object index 0.
 * All objects past the last valid objects should be invalid.
 * 
 * This code was implemented to be able to quickly search for the number of available memory blocks by polling which memory block is valid or invallid.
 */

#include <stdio.h>
#include <velib/base/base.h>

#include "sa_search.h" 

static un16 getCenterIndex (un16 iLow, un16 iHigh)
{
	return iLow + ((iHigh - iLow) >> 1);
}

/**
 * @brief initialize the saSearch state.
 *
 * @param state		pointer to saSearchState struct.
 * @param iMax		maximum index to inspect.
 */
void saSearchInit (SaSearchState* state, un16 iMax)
{
	state->i = 0;
	state->iLow = 0;
	state->iHinvalid = iMax;
	state->success = veFalse;
		
	state->iHnext = getCenterIndex(0, iMax);
}

/**
 * @brief Get the next index of the object to probe.
 *
 * With the first call to this function objectValidFlag should contain the restult of Object at index 0.
 * Use only after inititalisation with asSearchInit()
 * 
 * @param state				pointer to SaSearchState struct. 
 * @param objectValidFlag	result of last object validity probe. 
 * @return					veTrue if last valid object was reached.
 */
veBool saSearchGetNextIndex (SaSearchState* state, veBool objectValidFlag)
{
	if (state->i == 0) {
		if (objectValidFlag == veFalse) {
			state->success = veFalse;
			return veTrue;
		}

		state->i = state->iHinvalid; /* Set to max on init */
		
		return veFalse;
	}
	
	if (objectValidFlag) {
		/* TRUE */
		state->iLow = state->i;
		state->i = state->iHinvalid -1;

		state->iHnext = getCenterIndex(state->iLow, state->i); 

		if ((state->i == state->iLow)) {
			/* Search done state->i contains the index of the last valid object  */
			state->success = veTrue;
		}

	}
	else {
		/* FALSE */
		state->iHinvalid = state->i;
		state->i = state->iHnext;
		state->iHnext = getCenterIndex(state->iLow, state->iHnext);
	}

	if (state->i == 0) {			
		state->i = state->iHinvalid - 1;
		state->success = veTrue;
	}
		
	return state->success;
}

/**
 * @brief Return the number of valid objects.
 *
 * This should be called after saSearchGetNextIndex() returns veTrue 
 * 
 * @param state				pointer to SaSearchState struct. 
 * @return					the number of valid object.
 */
un16 saSearchReturnNValidObjects (SaSearchState* state)
{
	if (state->success)
		return state->i + 1;
	else
		return 0;
}


#ifdef SA_SEARCH_TEST
/**
 * @brief saSearch test code.
 *
 * Mode 0 = normal use test
 * Mode 1 = test with center gap. Just to test if it does not loop forever. Results are useless
 * Mode 2 - test with alternating gaps
 * 
 * @param mode				test mode. 
 */
void saSearchTest (un8 mode)
{
	SaSearchState state;
	veBool testObjects[256];
	veBool result;
	un16 i, n, nIter;
		
	
	for (i = 0; i < sizeof(testObjects); i ++) {
		testObjects[i] = veFalse;
	}


	for (n = 0; n < sizeof(testObjects); n ++) {
		printf ("-----------------------------------\n");
		printf ("Test with %d items. ", n);

		for (i = 0; i < n; i++) {
			testObjects[i] = veTrue;
		}

		switch (mode) {
			case 1:
				/* Create gap */
				printf("With center gap\n");
				for (i = (n >> 2); i < (n >> 1); i ++)
					testObjects[i] = veFalse;
				break;
			case 2:
				/* Create gaps every other object */
				printf("With alternating gaps\n");
				for (i = 1; i < n; i = i+12)
					testObjects[i] = veFalse;
				break;
			default:
				printf("Normal use\n");
		}
		
		saSearchInit(&state, 0xFF);

		nIter = 0;

		result = veFalse;
		
		while (result == veFalse) {
#ifdef SA_SEARCH_DEBUG
			printf("  Probing index %d", state.i);
			if (testObjects[state.i])
				printf(" (T)\n");
			else
				printf(" (F)\n");
#endif		
			result = saSearchGetNextIndex(&state, testObjects[state.i]);
			nIter ++;
			
			if (nIter >= 300)
				break;
		}

		if (result) {
			if (n != saSearchReturnNValidObjects(&state)) {
				if (mode == 0) {
					printf("FAILED wrong number of valid objects\n");
					printf("Number of objects = %d\n", saSearchReturnNValidObjects(&state));
					break;
				}
				else 
					printf("SUCCESS (Gap test)\n ");
				
			}
			else
				printf("SUCCESS\n");
		}
		else {
			printf("FAILED too many itterations\n");
			break;
		}
	
		printf("Number of itterations = %d\n", nIter);
	}
}
#endif





