#ifdef TASK_MK2RVSC

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include <velib/velib_inc_mk2.h>
#include <velib/base/base.h>
#include <velib/base/endian.h>
#include <velib/platform/console.h>
#include <velib/platform/plt.h>
#include <velib/platform/task.h>
#include <velib/mk2/frame_handler.h>
#include <velib/mk2/mk2.h>
#include <velib/mk2/hal.h>
#include <velib/mk2/mk2_actions.h>
#include <velib/mk2/mk2_strings.h>
#include <velib/mk2/mk2_update.h>
#include <velib/mk2/vebus_device.h>
#include <velib/mk2/vebus_items.h>
#include <velib/mk2/values_float.h>
#include <velib/utils/ve_logger.h>
#include <velib/base/ve_string.h>


#include "mk2vsc.h"
#include "vscFile.h"

#include "vll.h"
#include "vebusfr.h"
#include "task_vebus_configure.h"
#include "rvsc_dbus_service.h"
#include "mk2_init.h"
/**
 * Remote VEConfigure and VEBus configuration program
 */

/* device talking to */
static struct Mk2VscStateS ps;

static un8 startupTimer;
static un8 discoverTimer;
static un8 delayedExecTimer;
static un16 wdTimer = WD_TIMEOUT_TIME_50MS;
static veBool wdReset = veFalse;

static DelayedExecCallBack* delayedExecCb;

/* Function declarations */
void finishMainTask(int err);

void getExtraFlagAccessInfo(void);
void getSerialNumber (void);
void qeuryDone(void);
void vebusConfigCheckResult(int result);
void vebusConfigCheckDone(void);
void vebusConfigDone(int result);
void vebusResetDevicesDone (void);
void mk2DelayedClose (void);
void mk2ReOpenPort (void);
void rediscoverVebusDevices(void);
void initLongWinmonFrameSupport(void);
void startCommonTask(void);

void obtainNextSetting(void);
void setNextSetting(void);
void readVllCode(void);
void writeVllCodePart0(void);
void writeVllCodePart1(void);
void writeVllCodePart2(void);
void writeVllCodePart3(void);
void writeVllCodePart4(int result);
void writeVllCodePart5(void);
void writeVllCodeDone(void);
void queryNextSelectedDevice(void);
void multiResetDone(void);

/* This defines the default vsc file name. */
static void createVscFileName(char* s, VebusDevice* dev)
{
	char* ext;

	if (s[0] == '\0')
		ve_snprintf(s, MAX_PATH_LEN, "%u",dev->uniqueNumber);
	/* With the new discover dev->uniqueNumber is not always the master */

	if (strlen(s) > (MAX_PATH_LEN - 5))
		return;

	ext = strrchr(s, '.');		  /* Find extention */
	if (ext == NULL) {
		/* Add extention */
		if (ps.nVebusDevices == 1) {
			strcat(s,".rvsc");
		}
		else
			strcat(s,".rvms");
	}
}

/* Delayed execution functionality through taskTick */
void mk2vscDelayedExec(DelayedExecCallBack* cb, un8 n50msCount)
{
	delayedExecCb = cb;
	delayedExecTimer = n50msCount;
}

void mk2vscSetDefaults(void)
{
	ps.mode = READ_SETTINGS;
	ps.vscFile.pFile = NULL;
	ps.fileName[0] = '\0';		/* An empty fileName tells us to build a file name
									 * based on device model and id automatically.*/
}

void mk2vscSetMode(Mk2VscMode mode)
{
	ps.mode = mode;
}

Mk2VscMode mk2vscGetMode(void)
{
	return ps.mode;
}

void mk2vscSetFileName(char* name)
{
	strncpy(ps.fileName, name, MAX_PATH_LEN);
}

/* note: mk2 values need conversion to normal values, library functions can do this */
void mk2ActObtainedRawSettingCb(struct VebusDeviceS* dev, un8 id, un16 value)
{
	char buf[100];

	vebusFormatRawSettingValue(dev, id, value, buf, sizeof(buf));
	logI("settings", "obtained setting %d, %s, %s", id, vebusGetSettingName(id), buf);

	/* Write setting value to file */
	ps.pVscSettingDataSectionIn->settingData[id] = value;

	if (ps.mode == WRITE_SETTINGS) {
		/* Compare current setting with new setting.  */
		if (value != ps.pVscSettingDataSectionOut->settingData[id]) {
			ps.multiResetRequired = veTrue;
			mk2ActSetRawSetting(ps.vebusDevice, ps.iSetting);
		}
		else {
			/* No change */
			ps.iSetting ++;
			setNextSetting();
		}
		return;
	}

	ps.iSetting ++;
	obtainNextSetting();
}

/* note: mk2 values need conversion to normal values, library functions can do this */
void mk2ActObtainedRawVarCb(struct VebusDeviceS* dev, un8 id, un16 value)
{
	char buf[100];

	vebusFormatRawVariableValue(dev, id, value, buf, sizeof(buf));
	logI("values", "obtained value %d, %s, %s", id, vebusGetRamVarName(id), buf);
}

/* @retval the new value to set for id */
un16 mk2ActSetRawSettingCb(struct VebusDeviceS* dev, un8 id)
{
	VE_UNUSED(dev);

	return ps.pVscSettingDataSectionOut->settingData[id];
}

/* some action is done */
void mk2ActDoneCb(struct VebusDeviceS* dev, Mk2Actions action)
{
	VE_UNUSED(dev);
	VE_UNUSED(action);
}

/* Not used but required by vebus_system  */
void vebusAssistantFoundCallback(struct VebusDeviceS *dev, un8 assistentId, un8 ramVarOffset, un8 ramvarCount)
{
	VE_UNUSED(dev);
	VE_UNUSED(assistentId);
	VE_UNUSED(ramVarOffset);
	VE_UNUSED(ramvarCount);
}

/* confirmation that the settings is set */
void mk2ActSetSettingDoneCb(struct VebusDeviceS* dev, un8 id)
{
	VE_UNUSED(dev);

	/* Check if setting was successfull or password required */
	if(ps.vebusDevice->accessLevelRequired != 0) {
		/* Password should allready be unlocked. */
		finishMainTask(MK2VSC_ERR_PASSWORD);
	}

	ps.iSetting ++;
	setNextSetting();
}

void vebusCurrentLimitChanged(un8 acin, un16 limit)
{
	VE_UNUSED(acin);
	VE_UNUSED(limit);
}

/* trouble / disconnects etc */
void mk2NoResponseCallback(void)
{
	/*
	 * On embedded systems, this is a time to reinit and invalidate stored values.
	 * For the console tools, just exit with failure code.
	 */
	logE("mk2", "communication failure");

	if (++ps.restartCnt > N_COMM_FAIL_RESTART_MK2)
		finishMainTask(MK2VSC_ERR_MK2_COMM);

	ps.taskDoneCb = NULL;

	if (ps.multiVebusConfigCheck && (ps.mode == WRITE_SETTINGS)) {
		ps.multiVebusConfigCheck = veFalse;
		logE("task_mk2vsc", "Timeout while checking VE.Bus configuration");

		if (++ps.multiResetCnt > N_MULTI_RESET_CNT)
			finishMainTask(MK2VSC_ERR_VEBUS_SET_INC);

		rvscDBusSetState(MK2VSC_ST_RESETTING_VEBUS_PRODUCTS);
		vebusResetDevices(vebusResetDevicesDone);	/* Reset all Multies */
		return;
	}

	logE("task_mk2vsc", "restart the MK2");

	startupTimer = 30;
	ps.nVebusDevices = 0;
	mk2Init();

	vebusControlInit();
	vebusSystem.init.cb = NULL;
}

void mk2NewDeviceFound(struct VebusDeviceS* dev)
{
	logI("task_mk2rvsc", "new device found");
	ps.vebusDevices[ps.nVebusDevices++] = dev;
}

void startSetAllSettings(void)
{
#ifdef DEBUG
	vscFileSettingInfoSectionToLog (ps.pVscSettingInfoSectionOut);
	vscFileSettingDataSectionToLog (ps.pVscSettingDataSectionOut);
#endif

	/* Set all settings */
	ps.iSetting = 0;
	rvscDBusSetState(MK2VSC_ST_WRITE_SETTINGS_DATA);
	setNextSetting();
}

void sendPasswordHandler(void)
{
	if (ps.vebusDevice->accessLevelRequired != 0) {
		/* Wrong password */
		finishMainTask(MK2VSC_ERR_PASSWORD);
	}

	if (++ps.iSetting < ps.pVscSettingDataSectionOut->nPasswords) {
		/* Send next password */
		vebusSendAccessPassword(ps.vebusDevice, &ps.pVscSettingDataSectionOut->passwords[ps.iSetting].password[0],
								ps.pVscSettingDataSectionOut->passwords[ps.iSetting].length, sendPasswordHandler);
	}

	startSetAllSettings();
}

void startVeConfigTask(void)
{
	/* The VeBus configuration should be done by now
	 * Configure the individual Multi's.
	 * Note: VeBus setting that go to file are also read from the "VeConfig" task.
	 */
	int result;
	un64 currentTime;

	if (vebusSupportedSetting(ps.vebusDevice, VEBUS_GRID_COMPLIANCE_SETTINGS_START_INDEX))
		ps.vebusSettingsCount = VEBUS_MAX_SETTINGS_COUNT;
	else
		ps.vebusSettingsCount = VEBUS_GRID_COMPLIANCE_SETTINGS_START_INDEX;

	switch (ps.mode) {
		case READ_SETTINGS:
			if (ps.vscFile.pFile == NULL) {
				createVscFileName(&ps.fileName[0], ps.vebusDevice);

				result = vscFileOpen(&ps.vscFile, &ps.fileName[0], 'w');
				if (result != 0) {
					finishMainTask(result);
					break;
				}
			}

			/* Get all settings */
			ps.iSetting = 0;
			ps.iSettingLast = 0;
			rvscDBusSetState(MK2VSC_ST_READ_SETTING_DATA);
			obtainNextSetting();
			break;
		case WRITE_SETTINGS:
			result = vscFileLoadSettingInfo(&ps.vscFile, ps.pVscSettingInfoSectionOut, ps.pVscSettingDataSectionIn->firmwareVersionNumber);
			if (result != 0)
				finishMainTask(result);

			result = vscFileLoadSettingData(&ps.vscFile, ps.pVscSettingDataSectionOut, ps.pVscSettingDataSectionIn->uniqueTargetNumber);
			if (result != 0)
				finishMainTask(result);

			/* Check if short ID still maches the uniqueTargetNumber */
			if (ps.pVscSettingDataSectionIn->vebusSetting[VEBUS_SETTING_SHORT_ID_INDEX] !=
			ps.pVscSettingDataSectionOut->vebusSetting[VEBUS_SETTING_SHORT_ID_INDEX]){
				logE("mk2vsc", "Incompatible short ID, uniqueTargetNumber combination");
				finishMainTask(MK2VSC_ERR_VEBUS_SET_INC);
			}

			/* Check if if a password is included. */
			if (ps.pVscSettingDataSectionOut->nPasswords > 0) {
				/* Check if settings are still valid */
				currentTime = (un64) time (NULL);
				if (currentTime - ps.pVscSettingDataSectionOut->settingsTime > VSC_SF_SETTING_VALID_TIME_SEC)
					finishMainTask(MK2VSC_ERR_FILE_EXPIRED); /* Settings file too old */
				if ((ps.pVscSettingDataSectionOut->validTime < currentTime - VSC_SF_SETTING_VALID_TIME_SEC) ||
				(ps.pVscSettingDataSectionOut->validTime > currentTime + VSC_SF_SETTING_VALID_TIME_SEC))
					finishMainTask(MK2VSC_ERR_FILE_EXPIRED); /* Settings file too old */

				ps.iSetting = 0;
				/* Send all passwords. Unlocks are valid for 30 seconds. Sufficient time to set all settings */
				vebusSendAccessPassword(ps.vebusDevice, &ps.pVscSettingDataSectionOut->passwords[0].password[0],
										ps.pVscSettingDataSectionOut->passwords[0].length, sendPasswordHandler);
				return;
			}

			startSetAllSettings();

			break;
		default:
			finishMainTask(0);
	}
}

void finishMainTask(int err)
{
	if(err != 0)
		logI("mk2vsc","Exit code = %d", err);

	logI("mk2vsc","Done");

	if (ps.pVscSettingDataSectionIn != NULL)
		free(ps.pVscSettingDataSectionIn);
	if (ps.pVscSettingInfoSectionIn != NULL)
		free(ps.pVscSettingInfoSectionIn);
	if (ps.pVscSettingInfoSectionOut != NULL)
		free(ps.pVscSettingInfoSectionOut);
	if (ps.pVscSettingDataSectionOut != NULL)
		free(ps.pVscSettingDataSectionOut);

	mk2Close();

	pltExit(err);
}

void gotVeBusConfig(void)
{
	un8 i;

	logI("gotVeBusConfig", "uniqueNumber = 0x%x", ps.vebusDevice->uniqueNumber);
	logI("gotVeBusConfig", "type = %d", ps.vebusDevice->type);

	ps.pVscSettingDataSectionIn->uniqueTargetNumber = ps.vebusDevice->uniqueNumber;
	ps.pVscSettingDataSectionIn->type = ps.vebusDevice->type;
	ps.pVscSettingInfoSectionIn->type = ps.vebusDevice->type;

	for (i = 0; i < VSC_SF_N_VEBUS_SETTINGS; i++) {
		logI("gotVeBusConfig", "setting[%d] = 0x%x", i, ps.vebusDevice->config.raw[i]);
		ps.pVscSettingDataSectionIn->vebusSetting[i] = ps.vebusDevice->config.raw[i];
	}

	/* Completed all common tasks. Start actual read or write */
	startVeConfigTask();
}

void getExtraFlagAccessInfoHandler(void)
{
	if(!ps.vebusDevice->flagAccessInfo[ps.iSetting].valid) {
		getSerialNumber();		/* flagAccessInfo not supported. */
		return;
	}

	ps.iSetting++;
	if (ps.iSetting >= CFG_VEBUS_FLAG_REGISTER_COUNT) {
		getSerialNumber();		/* Done */
		return;
	}

	getExtraFlagAccessInfo();
}

void getExtraFlagAccessInfo(void)
{
	struct VebusDeviceS* dev;
	un8 settingId;
	un8 const flags[] = {VE_FLAGS_0, VE_FLAGS_1, VE_FLAGS_2, VE_FLAGS_3};

	dev = ps.vebusDevice;
	settingId = flags[ps.iSetting];
	vebusGetExtraFlagAccessInfo(dev, settingId, getExtraFlagAccessInfoHandler, &dev->flagAccessInfo[ps.iSetting]);
}

void getSerialNumber (void)
{
	vebusGetSerialNumber(ps.vebusDevice, qeuryDone, &ps.vebusDevice->serialNumber[0]);
}

/* done -> get ram / settings info */
void qeuryDone(void)
{
	un8 i, j, k;

	wdReset = veTrue;
	ps.pVscSettingDataSectionIn->firmwareVersionNumber = ps.vebusDevice->version.number;
	ps.pVscSettingInfoSectionIn->firmwareVersionNumber = ps.vebusDevice->version.number;

	/* Below settings are for future versions. */
	ps.pVscSettingDataSectionIn->compFirmwareNumber = ps.vebusDevice->version.number;
	ps.pVscSettingDataSectionIn->identification = 1;

	/* Add version frame remainder */
	ps.pVscSettingDataSectionIn->versionFrameRemainderLength = (un16) ps.vebusDevice->version.remainderLength;

	for (i = 0; i < VEBUS_VERSION_REMAINDER_LEN; i ++)
		ps.pVscSettingDataSectionIn->versionFrameRemainder[i] = ps.vebusDevice->version.remainder[i];

	/* Add settings info */
	for (i = 0; i < CFG_VEBUS_SETTINGS_COUNT; i++) {	/* SettingInfo in vsc file does not contain accessLevel */
		memcpy(&ps.pVscSettingInfoSectionIn->settingInfo[i], &ps.vebusDevice->settingInfo[i], sizeof(VscVebusSettingInfo));
		ps.pVscSettingInfoSectionIn->settingsAccessArray[i] = ps.vebusDevice->settingInfo[i].accessLevel;
	}

	k = 0;
	for (i = 0; i  < CFG_VEBUS_FLAG_REGISTER_COUNT; i ++) {
		if (ps.vebusDevice->flagAccessInfo[i].valid) {
			for (j = 0; j < 16; j ++)
				ps.pVscSettingInfoSectionIn->flagAccessArray[k++] = ps.vebusDevice->flagAccessInfo[i].accessLevel[j];
			ps.pVscSettingInfoSectionIn->flagAccessArrayCount = k;
		}

		else {
			/* Access info not supported. */
			ps.pVscSettingInfoSectionIn->flagAccessArrayCount = 0;
			break;
		}
	}

	j = 0;
	for (i = 0; i < VSC_SF_SERIAL_NUMBER_LEN; i ++) {
		j |= ps.vebusDevice->serialNumber[i];
		ps.pVscSettingDataSectionIn->serialNumber[i] = ps.vebusDevice->serialNumber[i];
	}

	if (j == 0)
		ps.pVscSettingDataSectionIn->serialNumberLength = 0;
	else
		ps.pVscSettingDataSectionIn->serialNumberLength = VSC_SF_SERIAL_NUMBER_LEN;

	/* Get the vebus settings info */
	rvscDBusSetState(MK2VSC_ST_READ_VEBUS_SETTINGS);
	vebusGetConfig(ps.vebusDevice, gotVeBusConfig);
}

void writeVscFile(void)
{
	un16 settingCount;
	int result;

	result = 0;

	/* Done create vsc file */
	settingCount = ps.iSettingLast + 1;
	ps.pVscSettingDataSectionIn->settingCount = settingCount;
	ps.pVscSettingInfoSectionIn->settingCount = settingCount;
	ps.pVscSettingInfoSectionIn->settingsAccessArrayCount = settingCount;

	/* Check if we allready have a settingInfo section. Otherwise write*/
	if (vscFileGetSettingInfoIndex(&ps.vscFile, ps.pVscSettingInfoSectionIn->firmwareVersionNumber, NULL) == veFalse)
		result = vscFileSaveSettingInfo(&ps.vscFile, ps.pVscSettingInfoSectionIn);

	if (result != 0)
		finishMainTask(result);

	if (vscFileGetSettingDataIndex(&ps.vscFile, ps.pVscSettingDataSectionIn->uniqueTargetNumber, NULL) == veFalse)
		result = vscFileSaveSettingData(&ps.vscFile, ps.pVscSettingDataSectionIn);

	if (result != 0)
		finishMainTask(result);


#ifdef DEBUG
	vscFileSettingInfoSectionToLog(ps.pVscSettingInfoSectionIn);
	vscFileSettingDataSectionToLog(ps.pVscSettingDataSectionIn);
#endif

	/* Check if we need to read more multi's */
	if (++ps.iDev == ps.nVebusDevices) {
		/* Done */
		vscFileClose(&ps.vscFile);
		finishMainTask(0);
	}

	/* Read next */
	queryNextSelectedDevice();
}

static void vllReadDoneCb(char * pBuf, un16 size, int result)
{
	VscVllSectionData* pVscVllSectionData = &ps.pVscSettingDataSectionIn->vllCode;

	VE_UNUSED(pBuf);

	pVscVllSectionData->size = size;

	logI("task_mk2vsc", "vllReadDoneCb. size = %d", size );
	wdReset = veTrue;

	if (result != 0) {
		logI("task_mk2vsc", "vll code read failed");
		pltExit(MK2VSC_ERR_VLL_READ);
	}

	if (ps.mode == READ_SETTINGS)
		writeVscFile();
	else if(ps.mode == WRITE_SETTINGS)
		writeVllCodePart0();
	else
		finishMainTask(0);
}

void mk2UpdateDone(Mk2Result result)
{
	if (result == MK2_UPDATE_FAILED) {
		logE("task_mk2vsc", "Incompatible MK2 firmware version");
		pltExit(MK2VSC_ERR_FIRMWARE_VERSION);
	}

	logI("task_mk2vsc", "Update done");
	startupTimer = 20;
}

void mk2InitDone(void)
{
	int i;

	logI("task_mk2vsc", "mk2InitDone");

	if (!vebusSystem.devices) {
		logE("task_mk2vsc", "No VEBus product was found");
		pltExit(MK2VSC_ERR_NO_VEBUS_PRODUCT);
		return;
	}

	if (ps.mode == LIST_PRODUCTS) {
		printf("Connected products: \n");
		for (i = 0; i < ps.nVebusDevices; i ++) {
			printf("  Short address = %d, ", ps.vebusDevices[i]->addr);
			printf("Unique number = 0x%08X\n", ps.vebusDevices[i]->uniqueNumber);
		}

		pltExit(0);
	}
	else {
		/* Check if stored vebus configuration maches the active configuration.
		 * After writing VE.Bus configuration the Multis should be reset.
		 * If this is not done the Multi can stay in a state where we can not reach the porper Multi by it's short address */
		ps.multiVebusConfigCheck = veTrue;
		vebusConfigCheck(&ps, vebusConfigCheckResult);
	}
}

void vebusConfigCheckResult(int result)
{
	if (result == 0) {
		ps.multiVebusConfigCheck = veFalse;
		vebusConfigCheckDone();
		return;
	}

	if (result == MK2VSC_ERR_VEBUS_SET_INC) {
		logE("task_mk2vsc", "VE.Bus config check failed. Reset required.");
		if (ps.mode == WRITE_SETTINGS)
		{
			if (++ps.multiResetCnt > N_MULTI_RESET_CNT)
				finishMainTask(result);

			ps.multiVebusConfigCheck = veFalse;
			rvscDBusSetState(MK2VSC_ST_RESETTING_VEBUS_PRODUCTS);
			vebusResetDevices(vebusResetDevicesDone);	/* Reset all Multies */
			return;
		}
	}
	finishMainTask(result);
}

void vebusConfigCheckDone(void)
{
	int result;

	logI("task_mk2vsc", "VE.Bus config check done");
	switch (ps.mode) {
		case READ_SETTINGS:
			/* check if all short addresses are unique, otherwise make them unique and reset */
			vebusConfigFixAddressConflicts(&ps, vebusConfigDone);
			break;
		case WRITE_SETTINGS:
			ps.pVscSettingInfoSectionOut = malloc(sizeof(VscSettingInfoSection));
			if (ps.pVscSettingInfoSectionOut == NULL) {
				logE("mk2vsc","malloc failed");
				finishMainTask(MK2VSC_ERR_MALLOC);
			}

			ps.pVscSettingDataSectionOut = malloc(sizeof(VscSettingDataSection));
			if (ps.pVscSettingDataSectionOut == NULL) {
				logE("mk2vsc","malloc failed");
				finishMainTask(MK2VSC_ERR_MALLOC);
			}

			createVscFileName(&ps.fileName[0], ps.vebusDevice);

			result = vscFileOpen(&ps.vscFile, &ps.fileName[0], 'r');
			if (result != 0)
				finishMainTask(result);

			/* Write the system configuration */
			vebusConfigWriteAll(&ps, vebusConfigDone);
			break;
		default:
			pltExit(0);
			break;
	}
}

void vebusConfigDone(int result)
{
	if (result != 0)
		finishMainTask(result);

	/* Check if reset is required */
	if(ps.multiResetRequired) {
		/* Reset only if the VebusConfiguration was changed */
		ps.multiResetRequired = veFalse;
		rvscDBusSetState(MK2VSC_ST_RESETTING_VEBUS_PRODUCTS);
		vebusResetDevices(vebusResetDevicesDone);	/* Reset all Multies */
		return;
	}

	initLongWinmonFrameSupport();
}

void vebusResetDevicesDone (void)
{
	logI("task_mk2vsc", "Reset devices done");
	mk2vscDelayedExec(mk2DelayedClose, 100);
}

void mk2DelayedClose (void)
{
	logI("task_mk2vsc", "Closing port");
	mk2Close();
	mk2vscDelayedExec(mk2ReOpenPort, 100);
}

void mk2ReOpenPort (void)
{
	const char *serialDevice;
	/* re-open port. After device reset the can cable MK2 loses it's power */
	serialDevice = pltGetSerialDevice();
	logI("task_mk2vsc", "re-opening port %s", serialDevice);
	if (!mk2Open(serialDevice)) {
		logE("task_mk2vsc", "cannot re-open '%s'", serialDevice);
		finishMainTask(MK2VSC_ERR_MK2_COMM);
	}
	logI("task_mk2vsc", "Allow 5 seconds for all devices to be rebooted");
	mk2vscDelayedExec(rediscoverVebusDevices, 100);
}

void rediscoverVebusDevices(void)
{
	/* Reinit vebus device administration*/
	ps.nVebusDevices = 0;
	mk2Start(mk2InitDone);
}

void initLongWinmonFrameSupport(void)
{
	mk2InitLongWinmonFrSupp(startCommonTask);
}

void startCommonTask(void)
{
	logI("task_mk2vsc", "start common task");
	wdReset = veTrue;
	ps.iDev = 0;
	ps.vebusDevice = ps.vebusDevices[ps.iDev];
	ps.iSetting = 0;

	ps.pVscSettingInfoSectionIn = malloc (sizeof(VscSettingInfoSection));
	if(ps.pVscSettingInfoSectionIn == NULL) {
		logE("mk2vsc","malloc failed");
		pltExit(MK2VSC_ERR_MALLOC);
	}

	ps.pVscSettingDataSectionIn = malloc (sizeof(VscSettingDataSection));
	if(ps.pVscSettingDataSectionIn == NULL) {
		logE("mk2vsc","malloc failed");
		pltExit(MK2VSC_ERR_MALLOC);
	}

	vscFileInitSettingInfoSection(ps.pVscSettingInfoSectionIn);
	vscFileInitSettingDataSection(ps.pVscSettingDataSectionIn);

	rvscDBusSetState(MK2VSC_ST_QUERY_DEVICE);
	vebusQueryDevice(ps.vebusDevice, VE_UNITS_VSC, getExtraFlagAccessInfo);
}

veBool powerControlAllowed(void)
{
	if (!vebusCanBeControlled())
		return veFalse;

	return veTrue;
}

/*== init / update / timer code to keep things going */
void taskInit(void)
{
	startupTimer = 0;
	discoverTimer = 0;
	delayedExecTimer = 0;

	ps.nVebusDevices = 0;
	ps.restartCnt = 0;
	ps.multiResetCnt = 0;
	ps.multiResetRequired = veFalse;
	ps.multiVebusConfigCheck = veFalse;
	ps.taskDoneCb = NULL;
	ps.pVscSettingInfoSectionIn = NULL;
	ps.pVscSettingInfoSectionOut = NULL;
	ps.pVscSettingDataSectionIn = NULL;
	ps.pVscSettingDataSectionOut = NULL;
	ps.taskDoneCb = NULL;

	vebusDeviceAllocInit();
	mk2FhInit();
	mk2Init();

	rvscDBusInit();
	rvscDBusSetState(MK2VSC_ST_INIT);

	mk2HalInit();
#if CFG_WITH_MK2_RESET
	mk2IoInit();
#endif
	mk2ActInit();

	vebusControlInit();

	mk2BreakClear();

#if CFG_WITH_MK2_RESET
	mk2HoldInReset(veFalse); /* power the mk2 */
	if (powerControlAllowed())
		mk2EnsureVebusPower(veTrue); /* power vebus */
#endif

	mk2StartWithFwCheck(mk2InitDone, mk2UpdateDone);
	wdReset = veTrue;
	pltInterruptEnable();
}

void taskUpdate(void)
{
	rvscDBusUpdate();

	mk2FhUpdate();
	if (ps.vebusDevice)
		mk2ActUpdate(ps.vebusDevice);
}

/* This called at a 50ms interval */
void taskTick(void)
{
#if CFG_WITH_MK2_UPDATE
	if (mk2UpdateBusy()) {
		mk2FirmwareTick();
		return;
	}
#endif

#if CFG_WITH_MK2_RESET
	if (startupTimer == 20) {
		mk2HoldInReset(veTrue);

		/* Power Ve.Bus but only if it needs to be on and no BMS etc control it */
		if (powerControlAllowed())
			mk2EnsureVebusPower(veTrue);
	}

	/*
	 * Get the mk2 out of reset before the startupTimer expires, so it has
	 * some time to startup.
	 */
	if (startupTimer == 5)
		mk2HoldInReset(veFalse);
#endif

	mk2FhTick();

	/* begin the startup procedure */
	if (startupTimer != 0 && --startupTimer==0)
		discoverTimer = 20 * 3;

	if (discoverTimer != 0 && --discoverTimer==0)
		mk2StartWithFwCheck(mk2InitDone, mk2UpdateDone);

	if (delayedExecTimer != 0 && --delayedExecTimer==0) {
		if (delayedExecCb != NULL)
			delayedExecCb();
	}

	/* Watchdog */
	if (--wdTimer == 0)
		finishMainTask(MK2VSC_ERR_TIMEOUT);

	if (wdReset) {
		wdReset = veFalse;
		wdTimer = WD_TIMEOUT_TIME_50MS;
	}
}

void obtainNextSetting(void)
{
	veBool result;

	result = veFalse;

	while ((ps.iSetting < ps.vebusSettingsCount) && (result == veFalse)) {
		result = vebusSupportedSetting(ps.vebusDevice, ps.iSetting);
		if(result == veTrue) {
			logI("settings", "Get Raw setting id = %d", ps.iSetting);

			mk2ActGetRawSetting(ps.vebusDevice, ps.iSetting);
			ps.iSettingLast = ps.iSetting;
		}
		else {
			logI("settings", "setting id = %d not supported", ps.iSetting);
			ps.iSetting ++;
		}
	}

	if (ps.iSetting >= ps.vebusSettingsCount) {
		/* Done reading settings  */
		readVllCode();
	}
}

void setNextSetting(void)
{
	veBool result;

	result = veFalse;

	while ((ps.iSetting < ps.vebusSettingsCount) && (result == veFalse)) {
		result = vebusSupportedSetting(ps.vebusDevice, ps.iSetting);
		if(result == veTrue) {
			logI("settings", "Set Raw setting id = %d", ps.iSetting);
			if(vebusSettingChangeRequiresReset(ps.vebusDevice->version, ps.iSetting)) {
				/* Read setting first in order to compare */
				mk2ActGetRawSetting(ps.vebusDevice, ps.iSetting);
				return;
			}
			mk2ActSetRawSetting(ps.vebusDevice, ps.iSetting);
		}
		else {
			logI("settings", "setting id = %d not supported", ps.iSetting);
			ps.iSetting ++;
		}
	}

	if (ps.iSetting >= ps.vebusSettingsCount) {
		/* Check if VLL code should be used */
		if (ps.pVscSettingDataSectionOut->vebusSetting[7] & IGNORE_VLL_CODE_BIT) {
			logI("task_mk2vsc", "VLL code not used, skip write");
			writeVllCodePart4(0);
		}
		else
			readVllCode();	/* Done, read VLL code to check if writing is required */
	}

}

void readVllCode(void)
{
	VscVllSectionData* pVscVllSD = &ps.pVscSettingDataSectionIn->vllCode;
	veBool vllCodeSupported, vllCodeNotUsed;

	/* Check if VLL code is supported */
	vllCodeSupported = vllCheckSupport(ps.vebusDevice);

	if (!vllCodeSupported) {
		logI("task_mk2vsc", "Vll code not supported");
		if (ps.mode == READ_SETTINGS)
			writeVscFile();
		else if (ps.mode == WRITE_SETTINGS)
			writeVllCodeDone();
		return;
	}

	/* check if VLL code is used */
	if ((ps.vebusDevice->config.raw[7] & IGNORE_VLL_CODE_BIT) != 0)
		vllCodeNotUsed = veTrue;
	else
		vllCodeNotUsed = veFalse;

	rvscDBusSetState(MK2VSC_ST_READ_VLL_CODE);

	vllRead(ps.vebusDevice, &pVscVllSD->data[0], vllReadDoneCb, vllCodeNotUsed);
}

void writeVllCodePart0(void)
{
	veBool result;
	VscVllSectionData* pVscVllSDIn = &ps.pVscSettingDataSectionIn->vllCode;
	VscVllSectionData* pVscVllSDOut = &ps.pVscSettingDataSectionOut->vllCode;

	/* Check if vll Code update is required by comparison */
	result = vllCodeCompare(&pVscVllSDIn->data[0], pVscVllSDIn->size, &pVscVllSDOut->data[0], pVscVllSDOut->size);
	if (result) {
		logI("task_mk2vsc", "VLL code not changed, skip write");
		writeVllCodeDone();
	}
	else {
		ps.retransmitCnt = 0;
		writeVllCodePart1();
	}
}

void writeVllCodePart1(void)
{
	ps.vebusDevice->config.raw[7] |= IGNORE_VLL_CODE_BIT;
	vebusFrSetConfig(ps.vebusDevice); /* VLL code will be reenabled after VEBus settings write */
	mk2vscDelayedExec(writeVllCodePart2, 5);
}

void writeVllCodePart2(void)
{
	/* Check if the ignore VLL code bit was properly set */
	vebusFrGetConfig(ps.vebusDevice, writeVllCodePart3);
}

void writeVllCodePart3(void)
{
	VscVllSectionData* pVscVllSD = &ps.pVscSettingDataSectionOut->vllCode;

	/* Check if the ignore VLL code bit was properly set */
	if ((ps.vebusDevice->config.raw[7] & IGNORE_VLL_CODE_BIT) != 0) {
		logI("task_mk2vsc", "vllIgnoreVllCode set");
		ps.multiResetRequired = veTrue;
		rvscDBusSetState(MK2VSC_ST_WRITE_VLL_CODE);
		ps.retransmitCnt = 0;
		vllWrite(ps.vebusDevice, &pVscVllSD->data[0], pVscVllSD->size, writeVllCodePart4);
	}
	else {
		if (ps.retransmitCnt >= N_VEBUSFR_RETRANSMITS) {
			logI("task_mk2vsc", "Retry set ignore VLL code bit");
			ps.retransmitCnt ++;
			writeVllCodePart1();
		}
		else {
			logE("task_mk2vsc", "Ignore VLL code bit set failed");
			finishMainTask(MK2VSC_ERR_VEBUS_SET_WRITE);
		}
	}
}

void writeVllCodePart4(int result)
{
	if (result != 0) {
		logE("task_mk2vsc","Write vll code failed");
		finishMainTask(MK2VSC_ERR_VLL_WRITE);
	}

	/* Clear the VEBus ignore VLL code bit. If required by the file. */
	if (ps.pVscSettingDataSectionOut->vebusSetting[7] & IGNORE_VLL_CODE_BIT)
		ps.vebusDevice->config.raw[7] |= IGNORE_VLL_CODE_BIT;
	else
		ps.vebusDevice->config.raw[7] &= ~IGNORE_VLL_CODE_BIT;

	vebusFrSetConfig(ps.vebusDevice);
	mk2vscDelayedExec(writeVllCodePart5, 5);
}

void writeVllCodePart5(void)
{
	un8 flag1, flag2;

	/* Check if the ignore VLL code bit was properly set */
	flag1 = ps.pVscSettingDataSectionOut->vebusSetting[7] &= IGNORE_VLL_CODE_BIT;
	flag2 = ps.vebusDevice->config.raw[7] &= IGNORE_VLL_CODE_BIT;

	if (flag1 == flag2) {
		logI("task_mk2vsc", "vllIgnoreVllCode reset");
		writeVllCodeDone();
	}
	else {
		if (ps.retransmitCnt >= N_VEBUSFR_RETRANSMITS) {
			logI("task_mk2vsc", "Retry reset ignore VLL code bit");
			ps.retransmitCnt ++;
			writeVllCodePart4(0);
		}
		else {
			logE("task_mk2vsc", "Reset VLL code bit set failed");
			finishMainTask(MK2VSC_ERR_VEBUS_SET_WRITE);
		}
	}
}

void writeVllCodeDone(void)
{
	logI("task_mk2vsc", "Write VLL code done");
	wdReset = veTrue;

	/* Multi's with firmware versions >= 308 return themselves if reset is required  */
	if (ps.vebusDevice->resetRequired)
		ps.multiResetRequired = veTrue;

	/* Check if we need to write more multi's */
	if (++ps.iDev == ps.nVebusDevices) {
		if (ps.multiResetRequired) {
			/* Reset only if we wrote VLL code or changed settings that require reset. */
			ps.multiResetRequired = veFalse;
			rvscDBusSetState(MK2VSC_ST_RESETTING_VEBUS_PRODUCTS);
			vebusResetDevices(NULL);	/* Reset all Multies */
			vebusResetDevices(NULL);	/* There is no feedback, so send it twice */
			mk2vscDelayedExec(multiResetDone, 60);
			return;
		}
		else
			finishMainTask(0);	/* Done */
	}

	/* Write next */
	queryNextSelectedDevice();
}

void queryNextSelectedDevice(void)
{
	veBool ready;

	/* Select an query device */
	ready = mk2ActUpdate(ps.vebusDevice);	/* Make sure pending requests are finished */

	if(!ready)
		mk2vscDelayedExec(queryNextSelectedDevice, 2);

	ps.vebusDevice = ps.vebusDevices[ps.iDev];
	vscFileInitSettingInfoSection(ps.pVscSettingInfoSectionIn);
	vscFileInitSettingDataSection(ps.pVscSettingDataSectionIn);

	rvscDBusSetState(MK2VSC_ST_QUERY_DEVICE);
	ps.iSetting = 0;
	vebusQueryDevice(ps.vebusDevice, VE_UNITS_VSC, getExtraFlagAccessInfo);
}

void multiResetDone(void)
{
	rvscDBusSetState(MK2VSC_ST_DONE);
	finishMainTask(0);
}
#endif
