#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <velib/velib_inc_mk2.h>
#include <velib/base/base.h>
#include <velib/utils/ve_assert.h>
#include <velib/utils/ve_logger.h>

#include "mk2vsc.h"
#include "vebusfr.h"
#include "rvsc_dbus_service.h"

typedef struct VebusConfigAddrConflictS {
	un8 n;
	un8 i;
	un8 addr[CFG_VEBUS_DEVICES];
	struct VebusDeviceS* devices[CFG_VEBUS_DEVICES];
} VebusConfigAddrConflict;

typedef struct VebusConfigCompS {
	un32 uniqueNumber;
	un8 type;
	VebusConfig config;
} VebusConfigComp;

/* Function declarations */
static void vebusConfigFixConflict(void);
static void vebusConfigFixConflictWriteStart(void);
static void vebusConfigFixConflictWrite(void);
static void vebusConfigFixConflictWriteCheck(void);
static void vebusConfigFixConflictWriteDone(void);
static void vebusConfigWriteStart(void);
static void vebusConfigWrite(void);
static void vebusConfigWriteCheck(void);
static void vebusConfigWriteDone(void);
static void vebusConfigWriteSelectNext(void);
static void vebusConfigCheckConfigPart1Done(void);

static struct Mk2VscStateS* ps;
static struct VebusConfigAddrConflictS aConfl;
static struct VebusConfigCompS configComp;

static void vebusConfigFixConflict(void)
{
	Mk2VscCallback* cb;

	if (aConfl.i < aConfl.n)
		vebusFrGetConfig(aConfl.devices[aConfl.i], vebusConfigFixConflictWriteStart);
	else {
		/* Done */
		logI("vebus_configure", "Fix VEBus address conflict done");
		cb = ps->taskDoneCb;
		if(cb != NULL) {
			ps->taskDoneCb = NULL;
			cb(0);
		}
	}
}

static void vebusConfigFixConflictWriteStart(void)
{
	ps->retransmitCnt = 0;
	ps->multiResetRequired = veTrue;
	rvscDBusSetState(MK2VSC_ST_WRITE_VEBUS_SETTINGS);

	vebusConfigFixConflictWrite();
}

static void vebusConfigFixConflictWrite(void)
{
	struct VebusDeviceS* dev;

	dev = aConfl.devices[aConfl.i];
	logI("vebus_configure", "Fix VEBus address conflict for 0x%x", dev->uniqueNumber);
	/* Replace the short ID */
	dev->config.raw[8] = aConfl.addr[aConfl.i];
	vebusFrSetConfig(dev);
	mk2vscDelayedExec(vebusConfigFixConflictWriteCheck, 5);
}

static void vebusConfigFixConflictWriteCheck(void)
{
	struct VebusDeviceS* dev;

	dev = aConfl.devices[aConfl.i];
	vebusFrGetConfig(dev, vebusConfigFixConflictWriteDone);
}

static void vebusConfigFixConflictWriteDone(void)
{
	Mk2VscCallback* cb;
	struct VebusDeviceS* dev;

	logI("vebus_configure", "Address confict fix, write Vebus config done");

	dev = aConfl.devices[aConfl.i];

	if (dev->config.raw[8] != aConfl.addr[aConfl.i]) {
		if (ps->retransmitCnt >= N_VEBUSFR_RETRANSMITS) {
			logI("vebus_configure", "Retry write VEBus settings");
			ps->retransmitCnt ++;
			vebusConfigFixConflictWrite();
		}
		else {
			logE("vebus_configure", "VEBus setting write failed");
			cb = ps->taskDoneCb;
			if(cb != NULL) {
				ps->taskDoneCb = NULL;
				cb(MK2VSC_ERR_VEBUS_SET_WRITE);
			}
			return;
		}
	}

	aConfl.i ++;				/* Select next */
	vebusConfigFixConflict();
}

static void vebusConfigWriteStart(void)
{
	Mk2VscCallback* cb;
	int result;

	result = vscFileLoadSettingData(&ps->vscFile, ps->pVscSettingDataSectionOut, ps->vebusDevice->uniqueNumber);
	if (result != 0) {
		cb = ps->taskDoneCb;
		if(cb != NULL) {
			ps->taskDoneCb = NULL;
			cb(result);
		}

		return;
	}

	ps->retransmitCnt = 0;

	/* Read vebus configuration and add to VebusDeviceS */
	vebusFrGetConfig(ps->vebusDevice, vebusConfigWrite);
}

static void vebusConfigWrite(void)
{
	int result;

	result = -1;

	if ((ps->vebusDevice->uniqueNumber == ps->pVscSettingDataSectionOut->uniqueTargetNumber) &&
	(ps->vebusDevice->type == ps->pVscSettingDataSectionOut->type))
		result = memcmp(&ps->vebusDevice->config.raw[0], &ps->pVscSettingDataSectionOut->vebusSetting[0], VSC_SF_N_VEBUS_SETTINGS);

	if (result == 0) {
		logI("vebus_configure", "Vebus config not changed, skip write");
		vebusConfigWriteSelectNext();
		return;
	}

	logI("vebus_configure", "Write Vebus config");
	ps->multiResetRequired = veTrue;
	rvscDBusSetState(MK2VSC_ST_WRITE_VEBUS_SETTINGS);

	ps->vebusDevice->uniqueNumber = ps->pVscSettingDataSectionOut->uniqueTargetNumber;
	ps->vebusDevice->type = ps->pVscSettingDataSectionOut->type;
	memcpy(&ps->vebusDevice->config.raw[0], &ps->pVscSettingDataSectionOut->vebusSetting[0], VSC_SF_N_VEBUS_SETTINGS);

	vebusFrSetConfig(ps->vebusDevice);
	mk2vscDelayedExec(vebusConfigWriteCheck, 5);
}

static void vebusConfigWriteCheck(void)
{
	vebusFrGetConfig(ps->vebusDevice, vebusConfigWriteDone);
}

static void vebusConfigWriteDone(void)
{
	Mk2VscCallback* cb;
	int result;

	logI("vebus_configure", "Write Vebus config done");

	result = -1;

	if ((ps->vebusDevice->uniqueNumber == ps->pVscSettingDataSectionOut->uniqueTargetNumber) &&
	(ps->vebusDevice->type == ps->pVscSettingDataSectionOut->type))
		result = memcmp(&ps->vebusDevice->config.raw[0], &ps->pVscSettingDataSectionOut->vebusSetting[0], VSC_SF_N_VEBUS_SETTINGS);

	if (result != 0) {
		if (ps->retransmitCnt >= N_VEBUSFR_RETRANSMITS) {
			logI("vebus_configure", "Retry write VEBus settings");
			ps->retransmitCnt ++;
			vebusConfigWrite();
		}
		else {
			logE("vebus_configure", "VEBus setting write failed");
			cb = ps->taskDoneCb;
			if(cb != NULL) {
				ps->taskDoneCb = NULL;
				cb(MK2VSC_ERR_VEBUS_SET_WRITE);
			}
			return;
		}
	}

	vebusConfigWriteSelectNext ();
}

static void vebusConfigWriteSelectNext(void)
{
	Mk2VscCallback* cb;

	/* Check if we need to write more multi's */
	if (++ps->iDev == ps->nVebusDevices) {
		cb = ps->taskDoneCb;
		if(cb != NULL) {
			ps->taskDoneCb = NULL;
			cb(0);
		}
		return;
	}

	/* Write next */
	ps->vebusDevice = ps->vebusDevices[ps->iDev];
	vebusConfigWriteStart();
}

static void vebusConfigCheckConfigPart2Done(void)
{
	/* Compare results */
	Mk2VscCallback* cb;
	un8 i;
	veBool result;

	cb = ps->taskDoneCb;
	result = veTrue;

	if(configComp.uniqueNumber != ps->vebusDevice->uniqueNumber)
		result = veFalse;

	if(configComp.type != ps->vebusDevice->type)
		result = veFalse;

	for (i = 0; i < VSC_SF_N_VEBUS_SETTINGS; i++) {
		if (configComp.config.raw[i] != ps->vebusDevice->config.raw[i])
			result = veFalse;
	}

	if (!result) {
		/* Settings don't mach. */
		if(cb != NULL) {
			ps->taskDoneCb = NULL;
			cb(MK2VSC_ERR_VEBUS_SET_INC);
		}
		return;
	}

	/* Check if we need to check more multi's */
	if (++ps->iDev == ps->nVebusDevices) {
		if(cb != NULL) {
			ps->taskDoneCb = NULL;
			cb(0);
		}
		return;
	}

	/* Check next */
	ps->vebusDevice = ps->vebusDevices[ps->iDev];
	vebusFrGetConfig(ps->vebusDevice, vebusConfigCheckConfigPart1Done);
}


static void vebusConfigCheckConfigPart1Done(void)
{
	un8 i;

	/* Save the results for comparison. */
	configComp.uniqueNumber = ps->vebusDevice->uniqueNumber;
	configComp.type = ps->vebusDevice->type;

	for (i = 0; i < VSC_SF_N_VEBUS_SETTINGS; i++)
		configComp.config.raw[i] = ps->vebusDevice->config.raw[i];

	/* Now read the vebus configuration using the MK2 command */
	mk2TalkToAddress(ps->vebusDevice->addr);
	vebusGetConfig(ps->vebusDevice, vebusConfigCheckConfigPart2Done);
}

/**
 * @brief Check and fix short address configuration conflicts
 *
 * Check if all short addresses are unique.
 * If not we need to assign unique id's and reset the Multi's
 *
 * @param progState			Pointer to the Mk2VscStateS struct
 * @param cb				Callback when done
 */
void vebusConfigFixAddressConflicts(Mk2VscState* progState, Mk2VscCallback* cb)
{
	struct VebusDeviceS* dev;
	un8 i, j, addr;
	un32 knownIds, id, idTemp;

	ps = progState;
	veAssert(ps->taskDoneCb == NULL);

	logI("vebus_configure", "Check and fix address conflicts");

	ps->taskDoneCb = cb;
	knownIds = 0;

	aConfl.n = 0;
	aConfl.i = 0;

	for (i = 0; i < ps->nVebusDevices; i ++) {
		dev = ps->vebusDevices[i];
		id = (1 << dev->addr);
		if (id & knownIds) /* Conflicting address */
			aConfl.devices[aConfl.n++] = dev;
		else
			knownIds |= id;
	}

	if (aConfl.n == 0) {
		logI("vebus_configure", "No address conflicts found");
		cb(0);       /* Done, no conflicts */
		return;
	}

	j = 0;
	id = (1 << (CFG_VEBUS_DEVICES - 1));
	for (i = 0; i < aConfl.n; i ++) {
		/* Assign free addresses to solve conflicts.   */
		/* Find next free address (highest to lowest). */

		while ((id & knownIds) && (j < CFG_VEBUS_DEVICES)) {
			id >>= 1;
			j++;
		}

		if (j >= CFG_VEBUS_DEVICES) {
			/* No free address found */
			cb(MK2VSC_ERR_VEBUS_ADDR_OVERFLOW);
			return;
		}

		knownIds |= id;
		idTemp = id;
		addr = 0;

		while (idTemp >>= 1)
			++addr;

		aConfl.addr[i] = addr;
	}

	/* Reconfigure the conflicting devices */
	vebusConfigFixConflict();

	return;
}

/**
 * @brief Write the vebus config to all devices
 *
 * Loop over n CFG_VEBUS_DEVICES.
 *   Read the matching (by uniqueNumber) vebusConfig from the vsc file
 *   Write the vebus settings to the device.
 *
 * If vebusConfig in one or more devices was changed,
 * progState->multiResetRequired will be set to veTrue
 * This changes ps->iDev.
 *
 * @param progState			Pointer to the Mk2VscStateS struct
 * @param cb				Callback when done
 */
void vebusConfigWriteAll(Mk2VscState* progState, Mk2VscCallback* cb)
{
	ps = progState;
	veAssert(ps->taskDoneCb == NULL);

	ps->taskDoneCb = cb;

	ps->iDev = 0;
	ps->vebusDevice = ps->vebusDevices[0];

	vebusConfigWriteStart();
}


/**
 * @brief Check the vebus configuration using the MK2 vebusGetConfig command
 *
 * After a VE.Bus configuration change, the Multis should be reset.
 * The Multi reset command does not give feedback. If reset fails the Multis still communicate using the old short ID.
 * The MK2 vebusGetConfig command uses short addressing. If all returned vebus conigurations mach, we know the reset was succesfull.
 * *
 * Loop over n CFG_VEBUS_DEVICES.
 *   Read the matching (by uniqueNumber) vebusConfig from the vsc file, and compare the settings.
 *
 * @param progState			Pointer to the Mk2VscStateS struct
 * @param cb				Callback when done
 */
void vebusConfigCheck(Mk2VscState* progState, Mk2VscCallback* cb)
{
	logI("vebus_configure", "VE.Bus configuration check");
	ps = progState;
	veAssert(ps->taskDoneCb == NULL);

	ps->taskDoneCb = cb;

	ps->iDev = 0;
	ps->vebusDevice = ps->vebusDevices[0];

	/* First read the vebus configuration using the VE.Bus frame command. */
	vebusFrGetConfig(ps->vebusDevice, vebusConfigCheckConfigPart1Done);
}
