/**
 * VLL code(assistants) low level read and write functions.
 * Flash blocks are 64 bytes. VE.Bus supports up to 15 bytes/frame.
 * A 64 byte flash block read or write requires 6 frames/subblocks (5*11 bytes and 1*9 bytes)
 *
 * Reading and writing is performed through WINMON commands.
 *
 * Commands:
 *  CommandVLLWriteAsPrevious   = 0x58;    // 1 parameter:  blocknr to write
 *  CommandVLLWriteSubBlock0    = 0x59;    // 12 parameters:   blocknr + 11*data
 *  CommandVLLWriteSubBlock1    = 0x5A;    // 12 parameters:   blocknr + 11*data
 *  CommandVLLWriteSubBlock2    = 0x5B;    // 12 parameters:   blocknr + 11*data
 *  CommandVLLWriteSubBlock3    = 0x5C;    // 12 parameters:   blocknr + 11*data
 *  CommandVLLWriteSubBlock4    = 0x5D;    // 12 parameters:   blocknr + 11*data
 *  CommandVLLWriteSubBlock5    = 0x5E;    // 12 parameters:   blocknr + 9*data
 *  CommandVLLRead              = 0x5F;    // 2 parameters  subblock(0..5) +1, blocknr
 *
 * Reponses:
 *  CommandReplyVLLRead         = 0x96;
 *  CommandConfirmOperation     = 0x97;
 *
 * CommandVLLWriteSubBlock0 should be send last. This triggers the flash write
 * The  CommandVLLWritexxxx commando’s respond with an CommandConfirmOperation
 *
 * Block writing of blocks with identical data (0xFF’s most likely) can be done with CommandVLLWriteAsPrevious.
 * This will write the specified block with the last received 64 byte data.
 *
 * Reading:
 *  The CommandVLLRead can respond with
 *
 *  CommandNotSupported:
 *     - if no extra parameter then this Multi does simply not support this command
 *     - if extra parameter is 2 then this indicates that the requested block is invalid (devices does not have that much space for VLLCode)
 *
 *  CommandReplyVLLRead:
 *     - If followed by databytes then this is the data requested
 *     - If no databytes follow (only possible when requesting subblock 0) then all 64 bytes of the requested block are 0xFF
 *
 **/

#include <velib/velib_inc_mk2.h>
#include <velib/utils/ve_assert.h>
#include <velib/mk2/hal.h>
#include <velib/mk2/mk2.h>
#include <velib/mk2/vebus_system.h>
#include <velib/utils/ve_logger.h>
#include <velib/mk2/frame_handler.h>
#include <velib/base/types.h>

#include "vebus_vll.h"
#include "sa_search.h"

/* Winmon commands */
#define	WM_CMD_VLL_WRITE_AS_PREV			0x58
#define	WM_CMD_VLL_WRITE_SUB_BLOCK_0		0x59
#define	WM_CMD_VLL_WRITE_SUB_BLOCK_1		0x5A
#define	WM_CMD_VLL_WRITE_SUB_BLOCK_2		0x5B
#define	WM_CMD_VLL_WRITE_SUB_BLOCK_3		0x5C
#define	WM_CMD_VLL_WRITE_SUB_BLOCK_4		0x5D
#define	WM_CMD_VLL_WRITE_SUB_BLOCK_5		0x5E
#define	WM_CMD_VLL_READ						0x5F

/* winmon responses */
#define	WM_RSP_VLL_READ						0x96
#define	WM_RSP_CONFIRM_OPER					0x97

/* winmon extra parameters */
#define WM_EP_VLL_READ_INVALID_BLOCK		0x2


static VllWriteState writeState;
static VllReadState readState;
static VllProbeState probeState;

/* Function declarations */
static veBool vebusVllWriteSubBlock(VllWriteState* ws, Mk2Callback* cb);
static Mk2EventReply vebusVllWriteHandler(Mk2Event ev, Mk2MsgRx* msg);
static Mk2EventReply vebusVllReadHandler(Mk2Event ev, Mk2MsgRx* msg);
static void vebusVllReadSubBlockCb(void);
static Mk2EventReply vebusVllProbeSubBlock0Handler(Mk2Event ev, Mk2MsgRx* msg);
static void vebusVllProbeSubBlock0Cb(void);


static veBool vebusVllWriteSubBlock(VllWriteState* ws, Mk2Callback* cb)
{
	un8 i, n, subBlockOffset, wmCmd;
	char* pData;
	Mk2MsgTx* msg;
	struct VebusDeviceS* dev;

	dev = ws->dev;
	wmCmd = ws->subBlockCmd;
	pData = ws->pData;

	msg = mk2MsgXYZ(dev->addr, wmCmd);
	if (!msg)
		return veFalse;

	if (wmCmd < WM_CMD_VLL_WRITE_SUB_BLOCK_0 || wmCmd > WM_CMD_VLL_WRITE_SUB_BLOCK_5)
		return veFalse;

	n = 11;

	if (wmCmd == WM_CMD_VLL_WRITE_SUB_BLOCK_5)
		n = 9;

	subBlockOffset = (wmCmd - WM_CMD_VLL_WRITE_SUB_BLOCK_0) * 11;

	msg->buf[3] = ws->blockNr;
	for (i = 0; i < n; i ++) {
		msg->buf[i+4] = pData[i+subBlockOffset];
	}

	msg->length = 15;

	msg->ctx = ws;
	msg->callback = cb;
	msg->handler = vebusVllWriteHandler;

	mk2FhQueueFrame(msg);
	return veTrue;
}


static Mk2EventReply vebusVllWriteHandler(Mk2Event ev, Mk2MsgRx* msg)
{
	Mk2Callback* cb;
	VllWriteState* ws;

	if (ev != MK2_FRAME)
		return MK2_CONTINUE;

	if (!mk2IsXYZNoData(msg, WM_RSP_CONFIRM_OPER))
		return MK2_CONTINUE;

	logI("vebus_vll", "++++ got vll write reply ++++");

	ws = (VllWriteState*) mk2.fh.msgOut->ctx;

	/* Write next block */
	if (ws->subBlockCmd  > WM_CMD_VLL_WRITE_SUB_BLOCK_0 && ws->subBlockCmd <= WM_CMD_VLL_WRITE_SUB_BLOCK_5)	{
		ws->subBlockCmd --;		/* We send the blocks backwards (block0 write triggers flash write) */

		cb = NULL;

		if (ws->subBlockCmd == WM_CMD_VLL_WRITE_SUB_BLOCK_0) {
			cb = ws->writeDoneCallback;
		}

		vebusVllWriteSubBlock(ws, cb);
	}
	return MK2_DONE;
}



static Mk2EventReply vllWriteAsPrevHandler(Mk2Event ev, Mk2MsgRx* msg)
{
	if (ev != MK2_FRAME)
		return MK2_CONTINUE;

	if (!mk2IsXYZNoData(msg, WM_RSP_CONFIRM_OPER))
		return MK2_CONTINUE;

	return MK2_DONE;
}


static veBool vllReadSubBlock(VllReadState* rs)
{
	Mk2MsgTx* msg;
	struct VebusDeviceS* dev;

	dev = rs->dev;

	msg = mk2MsgXYZ(dev->addr, WM_CMD_VLL_READ);

	if (!msg) {
		return veFalse;
	}

	logI("vebus_vll", "vllReadSubBlock %d",rs->subBlockNr);

	msg->buf[3] = rs->subBlockNr + 1;
	msg->buf[4] = rs->blockNr;
	msg->length = 5;

	msg->ctx = rs;
	msg->callback = vebusVllReadSubBlockCb;
	msg->handler = vebusVllReadHandler;

	mk2FhQueueFrame(msg);
	return veTrue;
}


static Mk2EventReply vebusVllReadHandler(Mk2Event ev, Mk2MsgRx* msg)
{
	un8 i, n, subBlockOffset;
	char* pData;
	VllReadState* rs;

	if (ev != MK2_FRAME)
		return MK2_CONTINUE;

	if (!mk2IsXYZNoData(msg, WM_RSP_UNKNOWN_COMMAND) && !mk2IsXYZNoData(msg, WM_RSP_VLL_READ))
		return MK2_CONTINUE;

	rs = (VllReadState*) mk2.fh.msgOut->ctx;
	pData = &rs->pBuf->data[0];

	logI("vebus_vll", "++++ got vll read reply ++++");

	if (msg->buf[2] == WM_RSP_UNKNOWN_COMMAND) {
		if (msg->length == 3) {
			rs->pBuf->result = CMD_NOT_SUPPORTED;
			return MK2_DONE;
		}
		else if ((msg->length >= 4) && (msg->buf[3] == WM_EP_VLL_READ_INVALID_BLOCK)) {
			rs->pBuf->result = INVALID_BLOCK;
			return MK2_DONE;
		}
	}
	else if (msg->buf[2] == WM_RSP_VLL_READ) {
		if (msg->length < 6 && rs->subBlockNr == 0) {
			rs->pBuf->result = FF_BLOCK;
			return MK2_DONE;
		}
		else if (msg->length == 11 + 3) {
			n = 11;

			subBlockOffset = rs->subBlockNr * 11;

			if (rs->subBlockNr >= 5)
				n = 9;

			for (i = 0; i < n; i ++) {
				pData[i+ subBlockOffset] = msg->buf[i+3];
			}

			if (n == 9) {
				rs->pBuf->result = SUCCESS;
			}
			else {
				rs->pBuf->result = BUSY;
				rs->subBlockNr ++;
				vllReadSubBlock(&readState);
			}
			return MK2_DONE;
		}
	}
	return MK2_CONTINUE;
}


static void vebusVllReadSubBlockCb(void)
{
	Mk2Callback* callback;

	if (readState.pBuf->result == BUSY)
		return;
	else {
		callback = readState.readDoneCallback;

		if (callback)
			callback();
	}

}

/* Test read sub block 0 in order to find out if the block is valid. */
static veBool vllProbeSubBlock0(VllProbeState* ps)
{
	Mk2MsgTx* msg;
	struct VebusDeviceS* dev;

	dev = ps->dev;

	msg = mk2MsgXYZ(dev->addr, WM_CMD_VLL_READ);

	if (!msg) {
		return veFalse;
	}

	msg->buf[3] = 0 + 1;			/* subBlock 0 */
	msg->buf[4] = ps->sasst.i;
	msg->length = 5;

	msg->ctx = ps;
	msg->callback = vebusVllProbeSubBlock0Cb;
	msg->handler = vebusVllProbeSubBlock0Handler;

	logI("vebus_vll", "Probing block %d", ps->sasst.i);

	mk2FhQueueFrame(msg);
	return veTrue;
}


static Mk2EventReply vebusVllProbeSubBlock0Handler(Mk2Event ev, Mk2MsgRx* msg)
{
	VllProbeState* ps;
	veBool result;

	if (ev != MK2_FRAME)
		return MK2_CONTINUE;

	ps = (VllProbeState*) mk2.fh.msgOut->ctx;

	if (!mk2IsXYZNoData(msg, WM_RSP_UNKNOWN_COMMAND) && !mk2IsXYZNoData(msg, WM_RSP_VLL_READ))
		return MK2_CONTINUE;

	logI("vebus_vll", "++++ got vll probe reply ++++");

	if((ps->sasst.i == ps->sasst.iLow) &&
	(ps->sasst.iHnext == ps->sasst.iHinvalid) &&
	(ps->sasst.i == ps->sasst.iHinvalid)) {
		/* A wrongly marked block caused a sa search in the wrong area.  */
		logE("vebus_vll", "VLL probe failed due to wrongly marked block");
		ps->result = FAILED;
		return (MK2_DONE);
	}

	if (msg->buf[2] == WM_RSP_UNKNOWN_COMMAND) {
		if (msg->length == 3)
			ps->result = CMD_NOT_SUPPORTED;
		else if ((msg->length >= 4) && (msg->buf[3] == WM_EP_VLL_READ_INVALID_BLOCK)) {
			logI("vebus_vll","Invalid block");
			result = saSearchGetNextIndex(&ps->sasst, veFalse);
			if (result)
				ps->result = SUCCESS;
			else {
				ps->result = BUSY;
				vllProbeSubBlock0(ps);
			}
		}
		return MK2_DONE;
	}
	else if (msg->buf[2] == WM_RSP_VLL_READ) {
		/* Valid data (FF block or normal data) */
		logI("vebus_vll","Valid block");
		result = saSearchGetNextIndex(&ps->sasst, veTrue);
		if (result)
			ps->result = SUCCESS;
		else {
			ps->result = BUSY;
			vllProbeSubBlock0(ps);
		}
		return MK2_DONE;
	}
	return MK2_CONTINUE;
}


static void vebusVllProbeSubBlock0Cb(void)
{
	VebusVllProbeCallback* callback;
	un16 n;

	if (probeState.result == BUSY)
		return;
	else {
		n = saSearchReturnNValidObjects(&probeState.sasst);
		* probeState.pNBlocks = n;

		callback = probeState.probeDoneCallback;

		if (callback)
			callback(probeState.result);
	}
}

/**
 * @brief Write a complete (64 bytes) byte VLL data block.
 *
 * When done the specified callback function will be called.
 * This issues the WM_CMD_VLL_WRITE_AS_PREV command without the need to resent the data.
 *
 * @param dev				vebus device.
 * @param cb				Callback when done.
 * @param blockNr			The block number to write.
 * @param pBuf				result
 * @return					veFalse if mk2Message generation failed.
 */
veBool vebusVllWriteBlock(struct VebusDeviceS* dev, Mk2Callback* cb, un8 blockNr, char* pData)
{

	logI("vebus_vll", "vebusVllWriteBlock nr %d" , blockNr);

	writeState.blockNr = blockNr;
	writeState.subBlockCmd = WM_CMD_VLL_WRITE_SUB_BLOCK_5;
	writeState.pData = pData;
	writeState.dev = dev;
	writeState.writeDoneCallback = cb;

	return vebusVllWriteSubBlock(&writeState, NULL);
}


/**
 * @brief Write the specified block with the last received 64 byte data.
 *
 * When done the specified callback function will be called.
 * This issues the WM_CMD_VLL_WRITE_AS_PREV command without the need to resent the data.
 *
 * @param dev				vebus device.
 * @param cb				Callback when done.
 * @param blockNr			The block number to write.
 * @param pBuf				result
 * @return					veFalse if mk2Message generation failed.
 */
veBool vebusVllWriteBlockAsPrevious(struct VebusDeviceS* dev, Mk2Callback* cb, un8 blockNr)
{
	Mk2MsgTx* msg = mk2MsgXYZ(dev->addr, WM_CMD_VLL_WRITE_AS_PREV);
	if (!msg) {
		return veFalse;
	}

	msg->buf[3] = blockNr;
	msg->length = 4;

	msg->ctx = NULL;
	msg->callback = cb;
	msg->handler = vllWriteAsPrevHandler;

	mk2FhQueueFrame(msg);
	return veTrue;
}


/**
 * @brief Start reading a VLL datablock.
 *
 * When done the result can be obtained from within the callback function through pBuf.
 * A VLL data block contains 64 bytes.
 *
 * @param dev				vebus device.
 * @param cb				Callback when done.
 * @param blockNr			The block number to read
 * @param pBuf				result
 * @return					veFalse if mk2Message generation failed.
 */
veBool vebusVllReadBlock(struct VebusDeviceS* dev, Mk2Callback* cb, un8 blockNr, VllReadBuffer* pBuf)
{
	readState.blockNr = blockNr;
	readState.subBlockNr = 0;
	readState.dev = dev;
	readState.readDoneCallback = cb;
	readState.pBuf = pBuf;
	pBuf->result = BUSY;

	logI("vebus_vll", "vebusVllReadBlock");
	logI("vebus_vll", "addr = 0x%x", readState.dev->addr);

	return vllReadSubBlock(&readState);
}


/**
 * @brief Start probing the number of available blocks.
 *
 * When done the result can be obtained from within the callback function through pNBlocks.
 * Only subBlocks0 are probed. Probing is based on succesive approximation.
 *
 * @param dev				vebus device.
 * @param cb				Callback when done.
 * @param pNBlocks			result
 * @return					veFalse if mk2Message generation failed.
 */
veBool vebusVllProbeNBlocks(struct VebusDeviceS* dev, VebusVllProbeCallback* cb, un16* pNBlocks)
{
	* pNBlocks = 0;

	probeState.pNBlocks = pNBlocks;
	probeState.dev = dev;
	probeState.probeDoneCallback = cb;

	saSearchInit(&probeState.sasst, MAX_VLL_BLOCK_NR);

	return vllProbeSubBlock0(&probeState);
}
