#include <string.h>

#include <velib/velib_inc_mk2.h>
#include <velib/utils/ve_assert.h>
#include <velib/mk2/hal.h>
#include <velib/mk2/mk2.h>
#include <velib/mk2/vebus_system.h>
#include <velib/utils/ve_logger.h>
#include <velib/mk2/frame_handler.h>
#include <velib/base/types.h>
#include <velib/base/endian_generic.h>

#include "vebusfr.h"

static VebusDevice *pVebusDevice;
static Mk2Callback *doneCb;
static un8 sFlagsExt0Backup;

/* OTTO move this to ve_endian_generic after testing */
void cpu_to_le_u32_tp(void* ptr, un32 val)
{
	un8 *p = (un8 *) ptr;
	p[0] = (un8)(val & 0xFF);
	p[1] = (un8)((val >> 8) & 0xFF);
	p[2] = (un8)((val >> 16) & 0xFF);
	p[3] = (un8)((val >> 24) & 0xFF);
}

/* This function swaps the high and low nible of a byte */
static un8 nibSwap(un8 x)
{
	return ((x & 0xF) << 4) | (x >> 4);
}

/* Calculate the vebus frame checksum */
static sn8 vebusFrCalcChecksum(un8* data, un8 len)
{
	un8 i;
	sn8 cs;

	cs = 0;

	for (i = 0; i < 5; i=i+2)
		cs -= nibSwap(data[i]);

	for (i = 0; i < len; i++)
		cs -= data[i];

	cs -= len;

	return cs;
}

static void vebusFrGetConfigDone(void)
{
	if (doneCb != NULL)
		doneCb();
}

static void vebusFrRestoreSFlagsExt0(void)
{
	if (mk2.state.sFlagsExt0 != sFlagsExt0Backup) {
		mk2.state.sFlagsExt0 = sFlagsExt0Backup;
		mk2SetState(&mk2.state, vebusFrGetConfigDone);
	}
	else
		vebusFrGetConfigDone();
}

static Mk2EventReply vebusFrGetConfigHandler(Mk2Event ev, Mk2MsgRx *msg)
{
	VebusDevice *dev;
	un8 type;

	if (ev != MK2_FRAME)
		return MK2_CONTINUE;

	dev = (VebusDevice*) mk2.fh.msgOut->ctx;

	if (msg->buf[0] != VEBUS_RSP && msg->length != 15)
		return MK2_CONTINUE;

	type = msg->buf[5];
	/* type should be either 2,3,4 or between 0x20 and 0x2F */
	if (type < 2)
		return MK2_CONTINUE;
	if ((type > 4) && (type < 0x20))
		return MK2_CONTINUE;
	if (type > 0x2F)
		return MK2_CONTINUE;

	logI("vebusfr_set", "++++ got vebus config settings read reply ++++");

	dev->uniqueNumber = le_u32_to_cpu_p(&msg->buf[1]);
	dev->type = type;

	memcpy(&dev->config.raw[0],&msg->buf[6],9);

	return MK2_DONE;
 }

static void vebusFrGetConfigSendCmd()
{
	Mk2MsgTx* msg;
	un32 uniNum = pVebusDevice->uniqueNumber;

	msg = mk2TxAlloc();
	if (!msg)
		return;

	logI("vebusfr_set", "---- get vebus config settings ----");
	msg->addChecksum = veFalse;

	msg->buf[0] = VEBUS_CMD; /* mk2 cmd */
	cpu_to_le_u32_tp(&msg->buf[1], uniNum);
	msg->buf[5] = GET_BUS_CONFIG;

	msg->buf[6] = vebusFrCalcChecksum(&msg->buf[0], 6);

	msg->length = 6;			/* data length without checksum  */

	msg->ctx = pVebusDevice;
	msg->callback = vebusFrRestoreSFlagsExt0;
	msg->handler = vebusFrGetConfigHandler;

	mk2FhQueueFrame(msg);
}

/**
 * @brief Get the VEBus config settings from the product (Multi).
 *
 * This temporarily configures the MK2 to forward config responses.
 *
 * @param dev				Pointer to the VebusDevice struct.
 * @param callback			Callback when done
 */
void vebusFrGetConfig(VebusDevice *dev, Mk2Callback *callback)
{
	pVebusDevice = dev;

	logI("vebusfr_set", "vebusFrGetVeBusSettings");

	doneCb = callback;
	sFlagsExt0Backup = mk2.state.sFlagsExt0;

	if ((mk2.state.sFlagsExt0 & MK2_S_FLAG_E0_FORWARD_ALL_CONFIG_RESP) == 0){
		/* Set the MK2 state to forward the config respose */
		mk2.state.sFlagsExt0 |= MK2_S_FLAG_E0_FORWARD_ALL_CONFIG_RESP;
		mk2SetState(&mk2.state, vebusFrGetConfigSendCmd);
	}
	else
		vebusFrGetConfigSendCmd();
}

/**
 * @brief Write the VEBus config settings to the product (Multi).
 *
 * There is no acknowledgement. Use vebusFrGetConfig() to check if the operation was succesfull.
 * Vebus settings become active after Multi reset.
 *
 * @param dev				Pointer to the VebusDevice struct.
 */
void vebusFrSetConfig(VebusDevice *dev)
{
	Mk2MsgTx* msg;
	un8 i;
	un32 uniNum = dev->uniqueNumber;

	msg = mk2TxAlloc();
	if (!msg)
		return;

	msg->addChecksum = veFalse;

	msg->buf[0] = VEBUS_CMD; /* mk2 cmd */
	cpu_to_le_u32_tp(&msg->buf[1], uniNum);
	msg->buf[5] =  WRITE_BUS_CONFIG;

	for (i = 0; i < 9; i++)
		msg->buf[i+6] = dev->config.raw[i];

	msg->buf[15] = vebusFrCalcChecksum(&msg->buf[0], 15);

	msg->length = 15;			/* data length without checksum  */

	mk2FhQueueFrame(msg);
}

/** Instruct all listening devices to send out an ident response.
 */
void vebusFrSendIdentRequest(Mk2Handler handler)
{
	Mk2MsgTx* msg;
	logI("mk2_vebus", "Sending ident request");

	msg = mk2TxAlloc();
	if (!msg)
		return;

	msg->addChecksum = veFalse;

	msg->buf[0] = VEBUS_CMD; /* mk2 cmd */
	cpu_to_le_u32_tp(&msg->buf[1], VEBUS_DEVICE_ADDRESS_BROADCAST);
	msg->buf[5] = IDENT_REQUEST;
	msg->buf[6] = vebusFrCalcChecksum(&msg->buf[0], 6);

	msg->length = 6;			/* data length without checksum  */

	msg->ctx = NULL;
	msg->callback = NULL;
	msg->handler = handler;

	mk2FhQueueFrame(msg);
}

/**
 * Set the ident behaviour for a specific device.
 *
 * @param uniqueNumber	The long address of the device.
 * @param sendIdent		\c TRUE, the device should reply to ident requests. \c FALSE, not.
 * @param handler		Mk2Handler function.
 */
void vebusFrSetIdentBehaviour(un32 uniqueNumber, veBool sendIdent, Mk2Handler* handler)
{
	Mk2MsgTx* msg;
	sendIdent = !sendIdent;

	logI("mk2_vebus", "Setting ident behaviour on device 0x%08X to %d", uniqueNumber, sendIdent);

	msg = mk2TxAlloc();
	if (!msg)
		return;

	msg->addChecksum = veFalse;

	msg->buf[0] = VEBUS_CMD; /* mk2 cmd */
	cpu_to_le_u32_tp(&msg->buf[1], uniqueNumber);
	msg->buf[5] = IDENT_DISABLE;
	msg->buf[6] = (un8) sendIdent;
	msg->buf[7] = vebusFrCalcChecksum(&msg->buf[0], 7);

	msg->length = 7;			/* data length without checksum  */

	msg->ctx = NULL;
	msg->callback = NULL;
	msg->handler = handler;

	mk2FhQueueFrame(msg);
}
