/**
 * @brief Discovers the devices and their addresses on VE.Bus.
 *
 * - The target address of the MK2 is first set to broadcast
 * - All devices are set to respond to IDENT requests
 *	- IDENT request is sent
 *	- found devices are set to not report again
 *	- above two steps are repeated till no more devices are found
 */

#include <velib/velib_inc_mk2.h>
#include <velib/utils/ve_assert.h>
#include <velib/mk2/hal.h>
#include <velib/mk2/mk2.h>
#include <velib/mk2/vebus_system.h>
#include <velib/utils/ve_logger.h>
#include <velib/base/types.h>
#include <velib/base/endian_generic.h>

#include "vebusfr.h"
#include "vebusfr_address_discovery.h"

/* The VE.Bus address to use when the actual address is not known, and for some other commands. */
#define MK2_DEFAULT_VEBUS_ADDRESS   0xFF
/* Vebus frames do not get acknowledged. So we need to send them multiple times. */
#define VEBUS_FRAME_RETRANSMIT_COUNT 1

typedef enum {
	VEBUS_DISCOVERY_IDLE,
	VEBUS_DISCOVERY_ONGOING,
} VebusDiscoveryStateId;

typedef struct VebusDiscoveryStateS
{
	VebusDiscoveryStateId id;
	un8 retransmitCount;
	Mk2Callback* doneCb;
	VebusSystem* vebusSyst;
} VebusDiscoveryState;

static VebusDiscoveryState state = {.id = VEBUS_DISCOVERY_IDLE, .retransmitCount = 0};

/* Function declarations */
static Mk2EventReply vebusDiscoverIdentHandler(Mk2Event ev, Mk2MsgRx *msg);


/* returns NULL if not found */
static VebusDevice* vebusDeviceFindByUniqeuNumber(VebusDevice* dev, un32 uniqueNumber)
{
	while (dev != NULL)	{
		if (dev->uniqueNumber == uniqueNumber)
			return dev;

		dev = dev->next;
	}

	return NULL;
}

static void vebusDeviceFreeAll(VebusDevice* dev)
{
	VebusDevice* devNext;

	while (dev != NULL)	{
		devNext = dev->next;
		vebusDeviceFree(dev);
		dev = devNext;
	}
}

/* Callback for settings all VE.Bus devices to respond to ident request. */
static Mk2EventReply vebusDiscoverSetIdentBAllHandler(Mk2Event ev, Mk2MsgRx *msg)
{
	if (ev == MK2_TX_SENT) {
		if (++state.retransmitCount <= VEBUS_FRAME_RETRANSMIT_COUNT) /* Send multiple times. Just to be sure */
			vebusFrSetIdentBehaviour(VEBUS_DEVICE_ADDRESS_BROADCAST, veTrue, vebusDiscoverSetIdentBAllHandler);
		else { /* start requesting all devices */
			state.retransmitCount = 0;
			vebusFrSendIdentRequest(vebusDiscoverIdentHandler);
		}

		return MK2_DONE;
	}
	else
		return MK2_CONTINUE;
}

/* Callback after single VE.Bus device respond to ident request disable. */
static Mk2EventReply vebusDiscoverSetIdentBSingleHandler(Mk2Event ev, Mk2MsgRx *msg)
{
	if (ev == MK2_TX_SENT) {
		vebusFrSendIdentRequest(vebusDiscoverIdentHandler);
		return MK2_DONE;
	}
	else
		return MK2_CONTINUE;
}


/* raw replies after ident request */
static Mk2EventReply vebusDiscoverIdentHandler(Mk2Event ev, Mk2MsgRx *msg)
{
	VebusDevice *dev;
	un32 uniqueNumber;
	un8 command;
	un8 addr;

	switch (ev)
	{
		case MK2_TIMEOUT:
			logI("vebusfr_address_discovery", "++++ got address discovery read reply timeout ++++");
			if (++state.retransmitCount <= VEBUS_FRAME_RETRANSMIT_COUNT)
				vebusFrSendIdentRequest(vebusDiscoverIdentHandler);
			else {
				/* Done */
				state.id = VEBUS_DISCOVERY_IDLE;

				if (state.doneCb != NULL)
					state.doneCb();
			}
			return MK2_DONE;

		case MK2_FRAME:
		{
			if (msg->length < 11)
				return MK2_CONTINUE;

			if (msg->buf[0] != VEBUS_RSP)
				return MK2_CONTINUE;

			logI("vebusfr_address_discovery", "++++ got address discovery read reply ++++");

			/* Add this device to the list of detected targets. */
			uniqueNumber = le_u32_to_cpu_p(&msg->buf[1]);
			command = msg->buf[5];

			/* Check if ident frame */
			switch (command){
				case IDENT_MULTI:
				case IDENT_COMPACT:
					break;
				default:
					return MK2_CONTINUE;
			}

			addr = msg->buf[6];

			/* Walk through the vebussystem.devices and add the device in case it's new */
			dev = state.vebusSyst->devices;
			dev = vebusDeviceFindByUniqeuNumber(dev, uniqueNumber);

			logI("vebusfr_address_discovery", "Found device: 0x%08X", uniqueNumber);

			if (dev == NULL) {
				/* New device found */
				if (!(dev = vebusDeviceAlloc()))
					return MK2_CONTINUE; /* This is an error */

				dev->uniqueNumber = uniqueNumber;
				dev->addr = addr;
				dev->next = state.vebusSyst->devices;
				state.vebusSyst->devices = dev;
				logI("vebusfr_address_discovery", "Added device: 0x%08X, shortAddr = %d", dev->uniqueNumber, dev->addr);
			}

			/* Single shot to stop reporting... */
			vebusFrSetIdentBehaviour(uniqueNumber, veFalse, vebusDiscoverSetIdentBSingleHandler);

			return MK2_DONE;
		}

		default:
			break;
	}

	return MK2_CONTINUE;
}

/** Determine how many VE.Bus devices are present in the system, and what their
 * addresses are.
 *
 * @details
 *	\ref vebus_vbmSystemDiscoveryCompleteEvent is called when successfully finished.
 *  The MK2 is marked as disconnected when no response is received.
 */
void vebusDiscoverSystem(VebusSystem* vebusSyst, Mk2Callback* cb)
{
	if (state.id != VEBUS_DISCOVERY_IDLE)
		return;

	logI("vebusfr_address_discovery", "Start vebus discovery");

	/* Assume no devices */
	/* The MK2 must be set to the default address so that it will forward the
	   replies (This command requires a response and will block the queue). */
	mk2TalkToAddress(MK2_DEFAULT_VEBUS_ADDRESS);

	/* Enable ident responses on all devices (send it twice, there is no response) */
	state.id = VEBUS_DISCOVERY_ONGOING;
	state.retransmitCount = 0;
	state.doneCb = cb;
	state.vebusSyst = vebusSyst;

	vebusDeviceFreeAll(vebusSyst->devices);
	vebusSyst->devices = NULL;

	vebusFrSetIdentBehaviour(VEBUS_DEVICE_ADDRESS_BROADCAST, veTrue, vebusDiscoverSetIdentBAllHandler);
}
