/**
 * VLL code(assistants) read and write functions.*
 */
#include <stdlib.h>

#include <velib/base/types.h>
#include <velib/utils/ve_logger.h>
#include <string.h>

#include "mk2vsc.h"
#include "vebus_vll.h"
#include "vll.h"
#include "vscFile.h"

#define VLL_BUF_FILL_VAL 0xFF	/* The vll data buffer should contain an exact number of VLL_BLOCK_SIZE blocks.
								 * If not this value will be used to add to the tail of the buffer */
typedef struct {
	un16 nBlocks;				/* The number of available blocks */
	un16 blockNr;
	un16 size;
	VllReadCallback* readDoneCallback;
	VllWriteCallback* writeDoneCallback;
	char* pCodeSize;				/* Pointer to the code size field */
	char* pCode;					/* Pointer to the code section */
	char* pVllSectionData;
	char *pVllRawData;
	char *pVllRawDataRp;			/* Read pointer */
} VllCodeBuffer;

static VllCodeBuffer vllCodeBuf;
static VllReadBuffer vllReadBuf;
static struct VebusDeviceS* vebusDev;

/* Function declarations */
static un16 vllDecompress (char* pVllSectionData, un16 size, char* pVllRawData);
static void vllProbeNBlocksDone (VllReadResult probeResult) ;
static void vllReadHandler (void);
static void vllWriteHandler (void);


/**
 * @brief Decompress vsc setting section file "VLLCode" section into raw VLL code
 *
 * @param pVllSectionData	Pointer to the VLL section data storage location
 * @param pVllRawData		Pointer to the raw decompressed VLL data
 * @return					Number of raw bytes
 */
static un16 vllDecompress (char* pVllSectionData, un16 size, char* pVllRawData)
{
	un16 codeSize;
	char fillVal;
	char* rp;
	char* wp;
	int i;

	wp = pVllRawData;
	rp = pVllSectionData;

	while (rp < (pVllSectionData + size)) {
		/* Code */
		rp = vscFileBufReadUn16(&codeSize, rp);

		for (i = 0; i < codeSize; i++)
			* wp ++ = * rp ++;

		/* Gap */
		fillVal = * rp ++;
		rp = vscFileBufReadUn16(&codeSize, rp);

		for (i = 0; i < codeSize; i ++)
			* wp ++ = fillVal;
	}

	return wp - pVllRawData;
}


static void vllProbeNBlocksDone (VllReadResult probeResult)
{
	VllReadCallback* callback;
	char* wp;
	int result;

	callback = vllCodeBuf.readDoneCallback;
	wp = vllCodeBuf.pVllSectionData;

	wp = vscFileBufWriteUn16(0,wp); /* CodeString0 size */
	* wp++ = 0xFF; /* FillValue */
	wp = vscFileBufWriteUn16((vllCodeBuf.nBlocks*VLL_BLOCK_SIZE), wp);	/* FillLength */

	if (probeResult == SUCCESS)
		result = 0;
	else
		result = -1;

	if(callback)
		callback(vllCodeBuf.pVllSectionData, wp - vllCodeBuf.pVllSectionData, result);
}


static void vllReadHandler (void)
{
	VllReadCallback* callback;
	un16 codeSize, gapSize;
	veBool done;
	char* pGap;

	done = veFalse;

	pGap = vllCodeBuf.pCode;
	vscFileBufReadUn16(&codeSize, vllCodeBuf.pCodeSize);
	vscFileBufReadUn16(&gapSize, pGap + sizeof(un8));

	switch(vllReadBuf.result) {
		case FF_BLOCK:
			/* Add gap */
			gapSize += VLL_BLOCK_SIZE;
			break;
		case SUCCESS:
			if (gapSize != 0) {
				/* Start new section. */
				vllCodeBuf.pCodeSize = vllCodeBuf.pCode + 3; /* Skip gap */
				vllCodeBuf.pCode = vscFileBufWriteUn16(0, vllCodeBuf.pCodeSize);
				codeSize = 0;
			}

			gapSize = 0;

			memcpy(vllCodeBuf.pCode, &vllReadBuf.data[0], VLL_BLOCK_SIZE);

			vllCodeBuf.pCode += VLL_BLOCK_SIZE;
			pGap = vllCodeBuf.pCode;

			codeSize += VLL_BLOCK_SIZE;
			vscFileBufWriteUn16(codeSize, vllCodeBuf.pCodeSize);
			break;
		case CMD_NOT_SUPPORTED:
		case INVALID_BLOCK:
		default:
			/* Done */
			done = veTrue;
			break;
	}

	/* Always add a gap section   */
	* pGap = 0xFF; /* FillValue */
	vscFileBufWriteUn16(gapSize, pGap + 1); /* FillLength */

	vllCodeBuf.blockNr++;
	if (vllCodeBuf.blockNr > MAX_VLL_BLOCK_NR) {
		logE("vllReadHandler", "Can not read past maximum block count");
		done = veTrue;
	}

	if(!done)
		vebusVllReadBlock(vebusDev, vllReadHandler, vllCodeBuf.blockNr, &vllReadBuf);
	else {
		callback = vllCodeBuf.readDoneCallback;

		if(callback)
			callback(vllCodeBuf.pVllSectionData, (vllCodeBuf.pCode + 3) - vllCodeBuf.pVllSectionData, 0);
	}
}

static void vllWriteHandler (void)
{
	VllWriteCallback* callback;
	char* rpp;
	char* rp;
	int result;

	vllCodeBuf.blockNr ++;

	if (vllCodeBuf.blockNr == vllCodeBuf.nBlocks) {
		/* Done */
		logI("vll","vllWrite Done");
		free (vllCodeBuf.pVllRawData);

		callback = vllCodeBuf.writeDoneCallback;

		if(callback)
			callback(0);

		return;
	}

	rpp = vllCodeBuf.pVllRawDataRp;
	rp = rpp + VLL_BLOCK_SIZE;
	vllCodeBuf.pVllRawDataRp = rp;

	if (vllCodeBuf.blockNr != 0) {
		/* Current block with previous block */
		result = memcmp(rp, rpp, VLL_BLOCK_SIZE);
		if (result == 0) {
			vebusVllWriteBlockAsPrevious(vebusDev,vllWriteHandler, vllCodeBuf.blockNr);
			return;
		}
	}

	vebusVllWriteBlock(vebusDev, vllWriteHandler, vllCodeBuf.blockNr, rp);
}


/**
 * @brief Read the complete VLL code block and add it to the pVllSectionData buffer.
 *
 * Data is formatted according to the vsc setting section file "VLLCode" section.
 * When done the specified callback function will be called.
 *
 * codeNotUsed Should be set to veTrue if the VE.Bus device info indicates that the VLL code is unused.
 * This triggers a probe for available VLL code space.
 *
 * @param dev				vebus device.
 * @param pVllSectionData	Pointer to the VLL section data storage location
 * @param cb				Callback when done.
 * @param codeNotused		If veTrue create an "empty" VLL Code section with the size of the available VLL code memory
 */
void vllRead (struct VebusDeviceS* dev, char* pVllSectionData, VllReadCallback* cb, veBool codeNotUsed)
{
	veBool result;

	vebusDev = dev;

	vllCodeBuf.pVllSectionData = pVllSectionData;
	vllCodeBuf.pCodeSize = pVllSectionData; /* codeString0 size field */
	vllCodeBuf.pCode = pVllSectionData + sizeof (un16);

	vllCodeBuf.readDoneCallback = cb;
	vllCodeBuf.blockNr = 0;

	if (codeNotUsed) {
		/* pBuf should be initialized as "empty" with the avaliable code size */
		vebusVllProbeNBlocks(vebusDev, vllProbeNBlocksDone, &vllCodeBuf.nBlocks);
		return;
	}

	vscFileBufWriteUn16(0, vllCodeBuf.pCodeSize);
	* vllCodeBuf.pCode = 0xFF; /* FillValue */
	vscFileBufWriteUn16(0, vllCodeBuf.pCode + 1); /* FillLength */

	result = vebusVllReadBlock(vebusDev, vllReadHandler, vllCodeBuf.blockNr, &vllReadBuf);

	if (!result) {
		/* return failed through callback */
		if(cb)
			cb(pVllSectionData, 0, -1);

	}
}


/**
 * @brief Write the complete VLL code block to the product (Multi).
 *
 * Data is formatted according to the vsc setting section file "VLLCode" section.
 * When done the specified callback function will be called.
 * Please note that prior to writing the VLL code should be disabled.
 *
 * @param dev				vebus device.
 * @param pVllSectionData	Pointer to the VLL section data location
 * @param size				VLL section data size in bytes
 * @param cb				Callback when done.
 */
void vllWrite (struct VebusDeviceS* dev, char* pVllSectionData, un16 size,  VllWriteCallback* cb)
{
	char* pRawDataBuf;
	char* wp;
	int i, tailCount;

	vebusDev = dev;

	vllCodeBuf.writeDoneCallback = cb;
	vllCodeBuf.blockNr = 0;

	pRawDataBuf = malloc(MAX_VLL_BLOCK_NR * VLL_BLOCK_SIZE);
	if(pRawDataBuf == NULL) {
		logE("mk2vsc","malloc failed");

		if(cb)
			cb(MK2VSC_ERR_MALLOC);

		return;
	}

	vllCodeBuf.pVllRawData = pRawDataBuf;
	vllCodeBuf.pVllRawDataRp = pRawDataBuf;

	vllCodeBuf.size = vllDecompress (pVllSectionData, size, pRawDataBuf);

	tailCount = vllCodeBuf.size % VLL_BLOCK_SIZE != 0;
	wp = pRawDataBuf + vllCodeBuf.size;

	for (i = 0; i < tailCount; i++) {
		/* Add missing data to the raw buffer in order to have an exact number of blocks */
		* wp ++ = VLL_BUF_FILL_VAL;
	}

	vllCodeBuf.size += tailCount;

	vllCodeBuf.nBlocks = vllCodeBuf.size / VLL_BLOCK_SIZE;

	logI("vll","vllWrite n blocks = %d", vllCodeBuf.nBlocks);

	vebusVllWriteBlock(dev, vllWriteHandler, vllCodeBuf.blockNr, pRawDataBuf);
}

/**
 * @brief Compare two VLL sections.
 *
 * We need to compare RAW unpacked VLL data since VEConfigure packs data differently.
 *
 * @param pVllSectionData1	Pointer to the VLL section data location
 * @param size1				VLL section data size in bytes
 * @param pVllSectionData2	Pointer to the VLL section data location
 * @param size2				VLL section data size in bytes
 * @return					veTrue if equal
 */
veBool vllCodeCompare (char* pVllSectionData1, un16 size1, char* pVllSectionData2, un16 size2)
{
	un16 rawSize1, rawSize2;
	char* pRawDataBuf1;
	char* pRawDataBuf2;
	veBool result;

	result = veFalse;

	pRawDataBuf1 = malloc(MAX_VLL_BLOCK_NR * VLL_BLOCK_SIZE);
	if(pRawDataBuf1 == NULL)
		return veFalse;			/* no need to cast an error */

	pRawDataBuf2 = malloc(MAX_VLL_BLOCK_NR * VLL_BLOCK_SIZE);
	if(pRawDataBuf2 == NULL) {
		free (pRawDataBuf1);
		return veFalse;			/* no need to cast an error */
	}

	rawSize1 = vllDecompress (pVllSectionData1, size1, pRawDataBuf1);
	rawSize2 = vllDecompress (pVllSectionData2, size2, pRawDataBuf2);

	if (rawSize1 != rawSize2)
		result = veFalse;
	else if (memcmp(pRawDataBuf1, pRawDataBuf2, size1) == 0)
		result = veTrue;

	free(pRawDataBuf1);
	free(pRawDataBuf2);

	return result;
}


/**
 * @brief Check if the "Multi" firmware version supports vll code.
 *
 * Firmware versions above xxyy204 support vll code.
 *
 * @param version			vebus device firmware version
 * @return					veTrue if firmware version supports vll code
 *
 */
veBool vllCheckSupport(struct VebusDeviceS* dev)
{
	if ((dev->version.number % 1000) <= 204)
		return veFalse;
	else if ((dev->version.features & VEBUS_VERION_F_COMBINED_FRAME) == 0)
		return veFalse;			/* Winmon frame mode does not suppport assistant info */
	else
		return veTrue;
}
