/**
 * vsc section file parser functions
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <velib/base/ve_string.h>
#include <velib/utils/ve_logger.h>

#include "mk2vsc.h"
#include "vscFile.h"
#include "vebus_vll.h"

#define min(A,B) ((A)<(B) ? (A) : (B))

/**
 * @brief Calculates the section checksum.
 *
 * @param pStart	section start address.
 * @param pEnd		section end address.
 *
 * @return			checksum
 */
static un32 calcSectionChecksum (char* pStart, char* pEnd)
{
	un32 checksum, i, j, n;
	char* rp;

	checksum = 0;
	j = 0;

	rp = pStart;

	n = pEnd - pStart;

	for ( i = 0; i <= n; i++)
	{
		checksum += ((un32)(* rp++) & 0xFF) << j;
		j += 8;
		j %= 32;
	}

	return checksum;
}

/**
 * @brief Search for a given section by name.
 *
 * @note If found the file pointer points to the beginning of the section
 *
 * @param pStrIdent			section name string.
 * @param pAddrSectBegin	Section begin address from where to start searching and found section address on success.
 * @param pAddrSectEnd		Section end address. 0 if section not found.
 * @param pFile				vsc file pointer.
 */
static void vscFileSeekSection(char* pStrIdent, un32* pAddrSectBegin, un32* pAddrSectEnd, VscFile* pVscFile)
{
	FILE* pFile;
	char buf[VSC_SF_MAX_ID_STR_LEN];
	un16 len;
	un32 addr, result;
	veBool vscFileEnd;

	pFile = pVscFile->pFile;

	* pAddrSectEnd = 0;
	vscFileEnd = veFalse;

	if ((* pAddrSectBegin == 0) && (feof(pFile) == 0))
	{
		/* Search from beginning of file, skip File identifier */
		result = fread(&len,sizeof(len),1,pFile);
		if (result != 1)
			return; /* Read failed */

		* pAddrSectBegin = len + sizeof(len);
	}

	fseek(pFile, (long int)(*pAddrSectBegin + pVscFile->startAddr), SEEK_SET);

	while (feof(pFile) == 0 && !vscFileEnd)
	{
		result = fread(&len,sizeof(len),1,pFile);
		if (result != 1)
			return; /* Read failed */

		if(len > VSC_SF_MAX_ID_STR_LEN)
		{
			/* Section id String too long */
			logE("vscFile", "Section id string too long");
			return;
		}

		result = fread(buf, len, 1, pFile);
		if (result != 1)
			return; /* Read failed */

		result = fread(&addr,sizeof(addr),1,pFile); /* Get the section end pointer */
		if (result != 1)
			return; /* Read failed */

		if(strncmp(pStrIdent, buf, len) != 0)
		{
			/* Section id string does not mach  */
			/* Move to the next section */
			fseek(pFile, (long int)(addr + pVscFile->startAddr), SEEK_SET);
			*pAddrSectBegin = addr;
			if ((addr + pVscFile->startAddr) >= pVscFile->endAddr)
				vscFileEnd = veTrue;
		}
		else
		{
			/* Section found */
			len += sizeof(len) + sizeof(addr);		   /* Rewind size */
			fseek(pFile,-((long int) len) , SEEK_CUR); /* Return to beginning of section */
			* pAddrSectEnd = addr;
			return;
		}
	}

	/* Section not found EOF */
	return;
}

static int vscFileWriteFileId (VscFile* pVscFile)
{
	FILE* pFile;
	char* pBuf;
	char* wp;
	un16 len;
	int result;

	pFile = pVscFile->pFile;

	/* Malloc a buffer */
	pBuf = malloc(VSC_SF_MAX_ID_STR_LEN + sizeof(un16));
	if (pBuf == NULL)
	{
		logE("vscFile", "Malloc setting file id buffer failed");
		return MK2VSC_ERR_MALLOC;
	}

	wp = pBuf;

	len = ve_snprintf(NULL, 0, VSC_SF_FILE_ID_STR); /* Determine the size of the file id string */

	/* Write  file id length */
	wp = vscFileBufWriteUn16(len, wp);

	/* Write file id string */
	ve_snprintf(wp, len+1, VSC_SF_FILE_ID_STR);
	wp += len;

	result = fwrite(pBuf, sizeof(char), (wp - pBuf), pFile);
	if (result != (wp - pBuf))
	{
		logE("vscFile","system configuration file write failed");
		free(pBuf);
		return MK2VSC_ERR_FILE_WRITE;
	}

	free(pBuf);
	return 0;
}

static char* vscFileBufInitSection(char* name, un32 version, char* wp, char** endPointer)
{
	un16 len;

	len = ve_snprintf(NULL, 0, name); /* Determine the size of the section id string */

	/* Write section id length */
	wp = vscFileBufWriteUn16(len, wp);

	/* Write section id string */
	ve_snprintf(wp, len+1, name);
	wp += len;

	/* Reserve 4 bytes for the EndPointer */
	*endPointer = wp;
	wp = vscFileBufWriteUn32(0, wp);

	/* Write section version number */
	wp = vscFileBufWriteUn32(version, wp);

	return wp;
}


/* Register endPointer,calculate checksum and write to file  */
static int vscFileWriteSection(FILE* pFile, char* wp, char* pBuf, char* endPointer, un32 fileOffset)
{
	int result;
	un32 checksum;

	/* Set section end pointer */
	/* Endpointer points past the checksum field */
	vscFileBufWriteUn32(fileOffset + ((wp - pBuf) + sizeof(un32)), endPointer);

	/* Calculate checksum */
	checksum = calcSectionChecksum (pBuf, (wp-1));
	wp = vscFileBufWriteUn32(checksum, wp);

	result = fwrite (pBuf, sizeof(char), (wp - pBuf), pFile);
	if (result != (wp - pBuf))
	{
		logE("vscFile","vsc file write failed");
		fclose(pFile);
		return MK2VSC_ERR_FILE_WRITE;
	}

	return 0;
}


static int vscFileSettingInfoGetFirmwareVersion (VscFile* pVscFile, un32 sStartAddr, un32* pVersion)
{
	FILE* pFile;
	char buf[4];
	un16 len;
	int result;

	pFile = pVscFile->pFile;

	fseek(pFile, (long int) sStartAddr, SEEK_SET);

	result = fread(&len, sizeof(len), 1, pFile);
	if (result != 1)
		return MK2VSC_ERR_FILE_READ;

	fseek(pFile,(long int) (len + 8), SEEK_CUR); /* skip section id, ep, sversion */
	result = fread(&buf, sizeof(un8), sizeof(buf), pFile);
	if (result != sizeof(buf))
		return MK2VSC_ERR_FILE_READ;

	vscFileBufReadUn32(pVersion, buf);

	fseek(pFile, (long int) sStartAddr, SEEK_SET);

	return 0;
}

static int vscFileSettingDataGetTargetNumber (VscFile* pVscFile, un32 sStartAddr, un32* pTargetNumber)
{
	FILE* pFile;
	char buf[4];
	un16 len;
	int result;

	pFile = pVscFile->pFile;

	fseek(pFile, (long int) sStartAddr, SEEK_SET);

	result = fread(&len, sizeof(len), 1, pFile);
	if (result != 1)
		return MK2VSC_ERR_FILE_READ;

	fseek(pFile,(long int) (len + 8), SEEK_CUR); /* skip section id, ep, sversion */
	result = fread(&buf, sizeof(un8), sizeof(buf), pFile);
	if (result != sizeof(buf))
		return MK2VSC_ERR_FILE_READ;

	vscFileBufReadUn32(pTargetNumber, buf);

	fseek(pFile, (long int) sStartAddr, SEEK_SET);

	return 0;
}


/* Scan the VSC file sections and build an index */
static int vscFileCreateIndex (VscFile* pVscFile)
{
	FILE *pFile;
	un16 i;
	un32 sStartAddr, sEndAddr, resUn32;
	int result;

	pFile = pVscFile->pFile;
	sStartAddr = 0;
	i = 0;

	fseek(pFile, (long int) pVscFile->startAddr, SEEK_SET);	/* Search from beginning of file */
	while (1)
	{
		vscFileSeekSection(VSC_SF_SECT_ID_STR_INFO, &sStartAddr, &sEndAddr, pVscFile);

		if(feof(pFile) != 0 || sEndAddr == 0)
			break; /* Section not found */

		pVscFile->index.nSettingInfo++;
		pVscFile->index.settingInfo[i].sStartAddr=sStartAddr;
		pVscFile->index.settingInfo[i].sEndAddr=sEndAddr;
		result = vscFileSettingInfoGetFirmwareVersion(pVscFile, sStartAddr, &resUn32);
		if (result != 0)
			return result;

		pVscFile->index.settingInfo[i].key.firmwareVersion = resUn32;
		i++;
		sStartAddr = sEndAddr;
	}

	sStartAddr = 0;
	i = 0;

	fseek(pFile, (long int) pVscFile->startAddr, SEEK_SET);	/* Search from beginning of file */
	while (1)
	{
		vscFileSeekSection(VSC_SF_SECT_ID_STR_DATA, &sStartAddr, &sEndAddr, pVscFile);

		if (feof(pFile) != 0 || sEndAddr == 0)
			break; /* Section not found */

		pVscFile->index.nSettingData++;
		pVscFile->index.settingData[i].sStartAddr=sStartAddr;
		pVscFile->index.settingData[i].sEndAddr=sEndAddr;
		result = vscFileSettingDataGetTargetNumber(pVscFile, sStartAddr, &resUn32);
		if (result != 0)
			return result;

		pVscFile->index.settingData[i].key.uniqueTargetNumber = resUn32;
		i++;
		sStartAddr = sEndAddr;
	}
	return 0;
}

/**
 * @brief Write a un16 to buffer wp and return the pointer to the next location
 *
 * @param val	value to be written.
 * @param wp	write pointer.
 *
 * @return		wp to next
 */
char* vscFileBufWriteUn16 (un16 val, char* wp)
{
	* wp++ = (char) (val & 0xFF);
	* wp++ = (char) (val >> 8);

	return wp;
}

/**
 * @brief Write a un32 to buffer wp and return the pointer to the next location
 *
 * @param val	value to be written.
 * @param wp	write pointer.
 *
 * @return		wp to next
 */
char* vscFileBufWriteUn32 (un32 val, char* wp)
{
	* wp++ = (char) (val & 0xFF);
	* wp++ = (char) (val >> 8);
	* wp++ = (char) (val >> 16);
	* wp++ = (char) (val >> 24);

	return wp;
}

/**
 * @brief Write a un64 to buffer wp and return the pointer to the next location
 *
 * @param val	value to be written.
 * @param wp	write pointer.
 *
 * @return		wp to next
 */
char* vscFileBufWriteUn64 (un64 val, char* wp)
{
	un8 i;

	* wp++ = (char) (val & 0xFF);

	for (i = 8; i <= 56; i = i + 8)
		* wp++ = (char) (val >> i);

	return wp;
}

/**
 * @brief Read an un16 from buffer rp and return the pointer to the next location
 *
 * @param pVal	pointer to resulting value location.
 * @param rp	read pointer.
 *
 * @return		rp to next
 */
char* vscFileBufReadUn16 (un16* pVal, char* rp)
{
	* pVal = (un16)(* rp++) & 0xFF;
	* pVal |= ((un16)(* rp++) & 0xFF) << 8;

	return rp;
}

/**
 * @brief Read an un32 from buffer rp and return the pointer to the next location
 *
 * @param pVal	pointer to resulting value location.
 * @param rp	read pointer.
 *
 * @return		rp to next
 */
char* vscFileBufReadUn32 (un32* pVal, char* rp)
{
	* pVal = (un32)(* rp++) & 0xFF;
	* pVal |= ((un32)(* rp++) & 0xFF) << 8;
	* pVal |= ((un32)(* rp++) & 0xFF) << 16;
	* pVal |= ((un32)(* rp++) & 0xFF) << 24;

	return rp;
}

/**
 * @brief Read an un64 from buffer rp and return the pointer to the next location
 *
 * @param pVal	pointer to resulting value location.
 * @param rp	read pointer.
 *
 * @return		rp to next
 */
char* vscFileBufReadUn64 (un64* pVal, char* rp)
{
	un8 i;

	* pVal = (un64)(* rp++) & 0xFF;

	for (i = 8; i <= 56; i = i + 8)
		* pVal |= ((un64)(* rp++) & 0xFF) << i;

	return rp;
}

/**
 * @brief Search for the settingInfo index by firmware version
 *
 * @param pVscFile			pointer to the VscFile handle struct
 * @param firmwareVersion	firmware version number
 * @param pIndex			pointer to the result field. NULL if not used
 *
 * @return					veTrue if section is available
 *							veFalse if section is not available
 */
veBool vscFileGetSettingInfoIndex (VscFile* pVscFile, un32 firmwareVersion, un16* pIndex)
{
	un16 i;
	veBool success;

	success = veFalse;

	for (i = 0; i < pVscFile->index.nSettingInfo; i++)
	{
		if ((pVscFile->index.settingInfo[i].key.firmwareVersion) == firmwareVersion)
		{
			success = veTrue;
			if (pIndex != NULL)
				* pIndex = i;
			break;
		}
	}
	return success;
}



/**
 * @brief Search for the settingData index by unique target number
 *
 * @param pVscFile			pointer to the VscFile handle struct
 * @param targetNumber		unique target number
 * @param pIndex			pointer to the result field. NULL if not used
 *
 * @return					veTrue if section is available
 *							veFalse if section is not available
 */
veBool vscFileGetSettingDataIndex (VscFile* pVscFile, un32 targetNumber, un16* pIndex)
{
	un16 i;
	veBool success;

	success = veFalse;

	for (i = 0; i < pVscFile->index.nSettingData; i++)
	{
		if ((pVscFile->index.settingData[i].key.uniqueTargetNumber) == targetNumber)
		{
			success = veTrue;
			if (pIndex != NULL)
				* pIndex = i;
			break;
		}
	}
	return success;
}

/**
 * @brief open a vsc section file
 *
 * @param pVscFile			pointer to the VscFile handle struct.
 * @param fileName			File name
 * @param mode				'r' = read, 'w' = write
 * @return					0 success
 * @return					!=0 error
 */
int vscFileOpen (VscFile* pVscFile, char* fileName, char mode)
{
	char* binMode;

	pVscFile->startAddr = 0;
	pVscFile->index.nSettingInfo = 0;
	pVscFile->index.nSettingData = 0;

	if(mode == 'w')
		binMode = "wb";
	else
		binMode = "rb";

	logI("vscFile", "Open %s, mode = %s", fileName, binMode);

	if (!(pVscFile->pFile = fopen(fileName, binMode)))
	{
		logE("vscFile","File open failed");
		return MK2VSC_ERR_FILE_OPEN;
	}

	if(mode == 'w')
		return vscFileWriteFileId(pVscFile);

	fseek(pVscFile->pFile, 0, SEEK_END); /* end of file */
	pVscFile->endAddr = (un32) ftell(pVscFile->pFile);
	fseek(pVscFile->pFile, 0, SEEK_SET); /* beginning of file */
	return vscFileCreateIndex(pVscFile);
}

void vscFileClose (VscFile* pVscFile)
{
	logI("vscFile", "Close vsc file");
	fclose (pVscFile->pFile);
}

int vscFileLoadSettingInfo (VscFile* pVscFile, VscSettingInfoSection* pSetInfo, un32 firmwareVersion)
{
	FILE* pFile;
	char* pBuf;
	char* rp;
	un32 sStartAddr, sEndAddr, sLen, checksumC, checksumR, version;
	un16 i, len, recordLen;
	int result, returnCode;

	pFile = pVscFile->pFile;
	returnCode = 0;

	if(!vscFileGetSettingInfoIndex(pVscFile, firmwareVersion, &i))
	{
		logE("vscFile", "Settings info section not found");
		return MK2VSC_ERR_FILE_SECTION;
	}

	sStartAddr = pVscFile->index.settingInfo[i].sStartAddr;
	sEndAddr = pVscFile->index.settingInfo[i].sEndAddr;

	pBuf = malloc (sizeof(VscSettingInfoSection)+sizeof(checksumC));
	if (pBuf == NULL)
	{
		logE("vscFile", "Malloc setting info section read buffer failed");
		return MK2VSC_ERR_MALLOC;
	}

	sLen = sEndAddr - sStartAddr;
	fseek(pFile, (long int)(sStartAddr + pVscFile->startAddr), SEEK_SET);
	result = fread(pBuf, sizeof(un8), sLen, pFile);
	if (result != sLen)
	{
		logE("vscFile","Read Failed, result = %d", result);
		returnCode = MK2VSC_ERR_FILE_READ;
		goto error;
	}

	/* pBuf now contains a complete section  */
	/* Check section checksum */

	rp = pBuf + sLen - sizeof(checksumC);

	vscFileBufReadUn32(&checksumR, rp--);

	checksumC = calcSectionChecksum(pBuf, rp);
	rp++;

	if (checksumC != checksumR)
	{
		logE("vscFile", "Setting info section checksum invalid");
		returnCode = MK2VSC_ERR_FILE_CHECKSUM;
		goto error;
	}

	rp = pBuf;
	rp = vscFileBufReadUn16(&len,rp);
	rp += len + sizeof(un32);					/* Skip section identifier and endpointer */

	rp = vscFileBufReadUn32(&version, rp);

	if (version < 1 || version > 2 )
	{
		logE("vscFile", "Setting info section version number invalid");
		returnCode = MK2VSC_ERR_FILE_VERSION;
		goto error;
	}

	rp = vscFileBufReadUn32(&pSetInfo->firmwareVersionNumber, rp);

	pSetInfo->type = *rp++;

	rp = vscFileBufReadUn16(&recordLen, rp); /* Read BSIRecordLen */

	pSetInfo->settingCount = recordLen / sizeof(VscVebusSettingInfo);

	len = sizeof(VscVebusSettingInfo);

	for (i = 0; i < pSetInfo->settingCount; i ++)
	{
		memcpy(&pSetInfo->settingInfo[i], rp ,len);
		rp += len;
	}

	if (version == 1) {
		pSetInfo->flagAccessArrayCount = 0;
		pSetInfo->settingsAccessArrayCount = 0;
	}
	else { /* version >= 2 */
		rp = vscFileBufReadUn16(&recordLen, rp); /* Read FlagAccessRecordLen */
		pSetInfo->flagAccessArrayCount = recordLen;
		memcpy(&pSetInfo->flagAccessArray[0], rp, recordLen);
		rp += recordLen;

		rp = vscFileBufReadUn16(&recordLen, rp); /* Read SettingsAccessRecordLen */
		pSetInfo->settingsAccessArrayCount = recordLen;
		memcpy(&pSetInfo->settingsAccessArray[0], rp, recordLen);
		rp += recordLen;
	}

error:
	free (pBuf);
	return returnCode;
}

int vscFileLoadSettingData (VscFile* pVscFile, VscSettingDataSection* pSetData, un32 targetNumber)
{
	FILE* pFile;
	char* pBuf;
	char* rp;
	un32 sStartAddr, sEndAddr, sLen, checksumC, checksumR, version;
	un16 i, len, bsdLen, resUn16, count;
	int result, returnCode;

	pFile = pVscFile->pFile;
	returnCode = 0;

	if(!vscFileGetSettingDataIndex(pVscFile, targetNumber, &i))
	{
		logE("vscFile", "Settings data section not found");
		return MK2VSC_ERR_FILE_SECTION;
	}

	sStartAddr = pVscFile->index.settingData[i].sStartAddr;
	sEndAddr = pVscFile->index.settingData[i].sEndAddr;

	pBuf = malloc (sizeof(VscSettingDataSection)+sizeof(checksumC));
	if (pBuf == NULL)
	{
		logE("vscFile", "Malloc setting data section read buffer failed");
		return MK2VSC_ERR_MALLOC;
	}

	sLen = sEndAddr - sStartAddr;
	fseek(pFile, (long int)(sStartAddr + pVscFile->startAddr), SEEK_SET);
	result = fread(pBuf, sizeof(un8), sLen, pFile);
	if (result != sLen)
	{
		logE("vscFile","Read Failed, result = %d", result);
		returnCode = MK2VSC_ERR_FILE_READ;
		goto error;
	}

	/* pBuf now contains a complete section  */
	/* Check section checksum */

	rp = pBuf + sLen - sizeof(checksumC);

	vscFileBufReadUn32(&checksumR, rp--);

	checksumC = calcSectionChecksum(pBuf, rp);
	rp++;

	if (checksumC != checksumR)
	{
		logE("vscFile", "Setting data section checksum invalid");
		returnCode = MK2VSC_ERR_FILE_CHECKSUM;
		goto error;
	}

	rp = pBuf;
	rp = vscFileBufReadUn16(&len,rp);
	rp += len + sizeof(un32);					/* Skip section identifier and endpointer */

	rp = vscFileBufReadUn32(&version, rp);

	if (version < 1 || version > 2 )
	{
		logE("vscFile", "Setting data section version number invalid");
		returnCode = MK2VSC_ERR_FILE_VERSION;
		goto error;
	}

	rp = vscFileBufReadUn32(&pSetData->uniqueTargetNumber, rp);

	rp = vscFileBufReadUn32(&pSetData->compFirmwareNumber, rp);

	pSetData->identification = *rp++;

	rp = vscFileBufReadUn32(&pSetData->firmwareVersionNumber, rp);

	rp = vscFileBufReadUn16(&len, rp);

	if (len > VEBUS_VERSION_REMAINDER_LEN)
	{
		logE("vscFile", "Version frame remainder too long");
		returnCode = MK2VSC_ERR_FILE_FORMAT;
		goto error;
	}

	pSetData->versionFrameRemainderLength = len;

	memcpy (&pSetData->versionFrameRemainder, rp, len);

	rp += len;

	pSetData->type = *rp++;

	memcpy (&pSetData->vebusSetting[0], rp, VSC_SF_N_VEBUS_SETTINGS);

	rp += VSC_SF_N_VEBUS_SETTINGS;

	if (version >= 2) {
		/* Serialnumber */
		rp = vscFileBufReadUn16(&len,rp);
		if (len > VSC_SF_SERIAL_NUMBER_LEN) {
			logE("vscFile", "Serial number too long");
			returnCode = MK2VSC_ERR_FILE_FORMAT;
			goto error;
		}

		pSetData->serialNumberLength = len;
		memcpy (&pSetData->serialNumber, rp, len);
		rp += len;

		/* Access password */
		rp = vscFileBufReadUn16(&count,rp);
		pSetData->nPasswords = count;

		if (count > VSC_SF_N_PASSWORDS) {
			logE("vscFile", "Invalid number of access passwords");
			returnCode = MK2VSC_ERR_FILE_FORMAT;
			goto error;
		}

		for (i = 0; i < count; i ++) {
			rp = vscFileBufReadUn16(&len,rp);
			if (len > VSC_SF_PASSWORD_LEN) {
				logE("vscFile", "Password too long");
				returnCode = MK2VSC_ERR_FILE_FORMAT;
				goto error;
			}

			pSetData->passwords[i].length = len;
			memcpy (&pSetData->passwords[i].password[0], rp, len);
			rp += len;
		}

		rp = vscFileBufReadUn64(&pSetData->validTime, rp);
		rp = vscFileBufReadUn64(&pSetData->settingsTime, rp);
	}
	else {
		/* Version 1: init as unused */
		pSetData->serialNumberLength = 0;
		pSetData->nPasswords = 0;
		for (i = 0; i < VSC_SF_N_PASSWORDS; i ++)
			pSetData->passwords[i].length = 0;

		pSetData->validTime = 0;
		pSetData->settingsTime = 0;
	}

	rp = vscFileBufReadUn16(&bsdLen, rp);

	pSetData->settingCount = bsdLen / sizeof(un16);

	for (i = 0; i < pSetData->settingCount; i ++)
	{
		rp = vscFileBufReadUn16(&resUn16, rp);
		pSetData->settingData[i] = resUn16;
	}

	/* VLL code */
	sLen =  sLen - sizeof(checksumC) - (rp - pBuf);

	if (sLen != 0)
	{
		memcpy (&pSetData->vllCode.data[0], rp, sLen);
		pSetData->vllCode.size = sLen;
	}
	else
		pSetData->vllCode.size = 0;

error:
	free (pBuf);
	return returnCode;
}


int vscFileSaveSettingInfo (VscFile* pVscFile, VscSettingInfoSection* pSetInfo)
{
	FILE* pFile;
	char* pBuf;
	char* wp;
	char* pSection0EndPointer;
	un32 checksum;
	un16 len, count, recordLen;
	int result;
	un32 fileOffset;
	un8 i;

	pFile = pVscFile->pFile;

	logI("vscFile", "Write setting info section");

	fseek(pFile, 0, SEEK_END); /* set fp to EOF */
	fileOffset = ((un32) ftell(pFile)) - pVscFile->startAddr;

	/* Malloc a buffer */
	pBuf = malloc (VSC_SF_MAX_ID_STR_LEN + sizeof(VscSettingInfoSection) + sizeof(checksum));
	if (pBuf == NULL)
	{
		logE("vscFile", "Malloc setting info section write buffer failed");
		return MK2VSC_ERR_MALLOC;
	}

	wp = pBuf;

	wp = vscFileBufInitSection(VSC_SF_SECT_ID_STR_INFO, VSC_SF_SECT_VERSION_INFO, wp, &pSection0EndPointer);

	/* Write settings section data */
	wp = vscFileBufWriteUn32(pSetInfo->firmwareVersionNumber, wp);

	*wp++ = pSetInfo->type;

	recordLen = pSetInfo->settingCount * sizeof(VscVebusSettingInfo);

	wp = vscFileBufWriteUn16(recordLen, wp); /* Write BSIRecordLen */

	len = sizeof(VscVebusSettingInfo);
	count = pSetInfo->settingCount;

	for (i = 0; i < count; i++)	{
		memcpy (wp,&pSetInfo->settingInfo[i], len);
		wp += len;
	}

	/* Version 2 data */
	recordLen = pSetInfo->flagAccessArrayCount;
	wp = vscFileBufWriteUn16(recordLen,wp); /* Write FlagAccessRecordLen */

	for (i = 0; i < recordLen; i++)
		*wp++ = pSetInfo->flagAccessArray[i];

	recordLen = pSetInfo->settingsAccessArrayCount;
	wp = vscFileBufWriteUn16(recordLen,wp); /* Write settingAccessRecordLen */

	for (i = 0; i < recordLen; i++)
		*wp++ = pSetInfo->settingsAccessArray[i];

	result = vscFileWriteSection(pFile, wp, pBuf, pSection0EndPointer, fileOffset);
	if (result == 0)
	{
		i = pVscFile->index.nSettingInfo++;
		pVscFile->index.settingInfo[i].sStartAddr=fileOffset;
		pVscFile->index.settingInfo[i].sEndAddr= ((un32) ftell(pFile)) - pVscFile->startAddr;
		pVscFile->index.settingInfo[i].key.firmwareVersion = pSetInfo->firmwareVersionNumber;
	}

	free(pBuf);
	return result;
}

int vscFileSaveSettingData (VscFile* pVscFile, VscSettingDataSection* pSetData)
{
	FILE* pFile;
	char* pBuf;
	char* wp;
	char* pSection0EndPointer;
	un32 checksum;
	un16 len, bsdLen;
	int result;
	un32 fileOffset;
	un8 i;

	pFile = pVscFile->pFile;

	logI("vscFile", "Write setting data section");

	fseek(pFile, 0, SEEK_END); /* set fp to EOF */
	fileOffset = ((un32) ftell(pFile)) - pVscFile->startAddr;

	/* Malloc a buffer */
	pBuf = malloc (VSC_SF_MAX_ID_STR_LEN + sizeof(VscSettingDataSection) + sizeof(checksum));
	if (pBuf == NULL)
	{
		logE("vscFile", "Malloc setting data section write buffer failed");
		return MK2VSC_ERR_MALLOC;
	}

	wp = pBuf;

	wp = vscFileBufInitSection(VSC_SF_SECT_ID_STR_DATA, VSC_SF_SECT_VERSION_DATA, wp, &pSection0EndPointer);

	/* Write settings section data */
	wp = vscFileBufWriteUn32(pSetData->uniqueTargetNumber, wp);

	wp = vscFileBufWriteUn32(pSetData->compFirmwareNumber, wp);

	*wp++ = pSetData->identification;

	wp = vscFileBufWriteUn32(pSetData->firmwareVersionNumber, wp);

	len = pSetData->versionFrameRemainderLength;

	/* Write  versionFrameRemainder length */
	wp = vscFileBufWriteUn16(len, wp);

	for (i = 0; i < len; i ++)
		*wp++ = pSetData->versionFrameRemainder[i];

	*wp++ = pSetData->type;

	for (i = 0; i < VSC_SF_N_VEBUS_SETTINGS; i++)
	{
		*wp++ = pSetData->vebusSetting[i];
	}

	len = pSetData->serialNumberLength;
	wp = vscFileBufWriteUn16(len, wp);

	for (i = 0; i < len; i ++)
		*wp++ = pSetData->serialNumber[i];

	wp = vscFileBufWriteUn16(0, wp);	/* Access password (nPassword = 0) */
	wp = vscFileBufWriteUn64(0, wp);	/* ValidTime */
	wp = vscFileBufWriteUn64((un64) time(NULL), wp); /* SettingsTime */

	bsdLen = pSetData->settingCount * sizeof(un16);

	wp = vscFileBufWriteUn16(bsdLen, wp);

	for (i = 0; i < pSetData->settingCount; i++)
		wp = vscFileBufWriteUn16(pSetData->settingData[i], wp);

	if (pSetData->vllCode.size != 0)
	{
		memcpy(wp, &pSetData->vllCode.data[0], pSetData->vllCode.size);
		wp += pSetData->vllCode.size;
	}

	result = vscFileWriteSection(pFile, wp, pBuf, pSection0EndPointer, fileOffset);
	if (result == 0)
	{
		i = pVscFile->index.nSettingData++;
		pVscFile->index.settingData[i].sStartAddr=fileOffset;
		pVscFile->index.settingData[i].sEndAddr= ((un32) ftell(pFile)) - pVscFile->startAddr;
		pVscFile->index.settingData[i].key.uniqueTargetNumber = pSetData->uniqueTargetNumber;
	}

	free(pBuf);
	return result;
}

/**
 * @brief Initialize the VscSettingInfoSection buffer with default values
 *
 * @param pSetInfo	pointer to the destination buffer.
 */
void vscFileInitSettingInfoSection (VscSettingInfoSection* pSetInfo)
{
	un8 i;

	if (pSetInfo == NULL)
		return;

	pSetInfo->firmwareVersionNumber = 0;

	pSetInfo->type = 1;

	pSetInfo->settingCount = CFG_VEBUS_SETTINGS_COUNT;

	for (i = 0; i < CFG_VEBUS_SETTINGS_COUNT; i ++)
	{
		pSetInfo->settingInfo[i].scale = 0;
		pSetInfo->settingInfo[i].offset = 0;
		pSetInfo->settingInfo[i].defaultValue = 0;
		pSetInfo->settingInfo[i].minimum = 0;
		pSetInfo->settingInfo[i].maximum = 0;
		pSetInfo->settingsAccessArray[i] = 0;
	}

	pSetInfo->flagAccessArrayCount = 0;
	pSetInfo->settingsAccessArrayCount = 0;
}

/**
 * @brief Initialize the VscSettingDataSection buffer with default values
 *
 * @param pSetData	pointer to the destination buffer.
 */
void vscFileInitSettingDataSection (VscSettingDataSection* pSetData)
{
	un8 i;

	if (pSetData == NULL)
		return;

	pSetData->firmwareVersionNumber = 0;
	pSetData->versionFrameRemainder[0] = '\0';

	pSetData->uniqueTargetNumber = 0;

	pSetData->type = 1;

	for (i = 0; i < VSC_SF_N_VEBUS_SETTINGS; i ++)
		pSetData->vebusSetting[i] = 0;


	pSetData->serialNumberLength = 0;
	pSetData->nPasswords = 0;
	pSetData->validTime = 0;
	pSetData->settingsTime = 0;

	pSetData->settingCount = CFG_VEBUS_SETTINGS_COUNT;

	for (i = 0; i < CFG_VEBUS_SETTINGS_COUNT; i ++)
		pSetData->settingData[i] = 0;

	pSetData->vllCode.size = 0;
}


/**
 * @brief Debug print VscSettingSectionInfo buffer contents
 *
 * @param pSetInfo	pointer to the source buffer.
 */
void vscFileSettingInfoSectionToLog (VscSettingInfoSection* pSetInfo)
{
	int i;

	logI("vscFileBSI","firmwareVersionNumber = %d", pSetInfo->firmwareVersionNumber);
	logI("vscFileBSI","type = %d", pSetInfo->type);

	for (i = 0; i < pSetInfo->settingCount; i ++)
	{
		logI("vscFileBSI","settingInfo[%d]", i);
		logI("vscFileBSI","  scale = %d", pSetInfo->settingInfo[i].scale);
		logI("vscFileBSI","  offset = %d", pSetInfo->settingInfo[i].offset);
		logI("vscFileBSI","  defaultValue = %d", pSetInfo->settingInfo[i].defaultValue);
		logI("vscFileBSI","  minimum = %d", pSetInfo->settingInfo[i].minimum);
		logI("vscFileBSI","  maximum = %d", pSetInfo->settingInfo[i].maximum);
	}

	for (i = 0; i < pSetInfo->flagAccessArrayCount; i++)
		logI("vscFileBSI","flagAccessArray[%d] = 0x%x", i, pSetInfo->flagAccessArray[i]);


	for (i = 0; i < pSetInfo->settingsAccessArrayCount; i++)
		logI("vscFileBSI","settingsAccessArray[%d] = 0x%x", i, pSetInfo->settingsAccessArray[i]);
}

/**
 * @brief Debug print VscSettingSectionData buffer contents
 *
 * @param pSetData	pointer to the source buffer.
 */
void vscFileSettingDataSectionToLog (VscSettingDataSection* pSetData)
{
	int i;
	un16 len;

	logI("vscFileBSD","firmwareVersionNumber = %d", pSetData->firmwareVersionNumber);

	len = pSetData->versionFrameRemainderLength;
	logI("vscFileBSD","versionFrameRemainderLength = %d", len);

	for (i = 0; i < len; i ++)
		logI("vscFileBSD","versionFrameRemainder[%d] = 0x%x", i, pSetData->versionFrameRemainder[i]);

	logI("vscFileBSD","uniqueTargetNumber = 0x%x", pSetData->uniqueTargetNumber);
	logI("vscFileBSD","type = %d", pSetData->type);

	for (i = 0; i < VSC_SF_N_VEBUS_SETTINGS; i ++)
		logI("vscFileBSD","vebusSetting[%d] = 0x%x", i, pSetData->vebusSetting[i]);

	logI("vscFileBSD", "nPasswords = %d", pSetData->nPasswords);

	logI("vscFileBSD", "validTime = %ld", pSetData->validTime);
	logI("vscFileBSD", "settingsTime = %ld", pSetData->settingsTime);

	for (i = 0; i < pSetData->settingCount; i ++)
		logI("vscFileBSD", "settingData[%d] = 0x%x",i, pSetData->settingData[i]);

	logI("vscFileBSD","VLL Code size = %d", pSetData->vllCode.size);
}
